c Subroutine to perform model selection on y_lam=x beta + epsi_delta
c using Monte Carlo Markov Chain techniques. Programmed by Michael Smith
c at the Australian Grad. School of Management, UNSW, Sydney
c v1 programmed 12/93; v2-3 1-8/94; Release v4.1 7-8/95 
c release v4.2 2/96

	subroutine bms4_2(n,p,txx,txy,tyy,indic,jay,
     +  it,Rone,Rzero,order,sens,indicprior,random,seed,lam,
     +  lnjacob,lamprior,trans,beta,vble,x,y,b,delta,
     +  outlier,deltaprior,deltamix,aa,bb,type)

	integer n,p,indic(*),jay(*),it,delta(2,*)
	integer order(*),lam,vble(*),type
	double precision seed,lnjacob(*),deltamix(*),b,x(n,*)
	double precision sens,indicprior(*),lamprior(*),aa(*),bb(*)
	double precision txy(9,*),txx(it,*),tyy(*),beta(*)
	double precision Rone(it,*),Rzero(it,*),y(*),deltaprior(*)
	logical random,trans,outlier

c indic:indicator variables (indic(i)=1 if beta(i)!=0)
c n:#of obns, p:# colms of design matrix x, jay:length of gibbs stages
c txx,txy,tyy are x'x, x'y, y'y 
c indicprior(i):prior for indic(i)=1 
c delta(1.) old vector of outlier indicators
c delta(2.) new vector of outlier indicators

c local variables

	integer i,j,k,gmaxp,q,l,sumdelta
	parameter(gmaxp=450)
	integer perm0(gmaxp),perm1(gmaxp),iwk(gmaxp)
	double precision work(gmaxp),tvect(gmaxp),betahat(gmaxp)
	double precision sse1(9),sse0(9),mixturelam(9),kk(gmaxp)
	double precision wk2(gmaxp)
	logical mix,gdbug1,gdbug2,gpflag,dump
	parameter(gdbug1=.false.,gdbug2=.false.,gpflag=.false.)
	parameter(dump=.false.)

	call standardisex(n,p,kk,x,it)
	call initialise(n,p,q,mix,indic,it,txx,Rone,Rzero,
     +  sse0,sse1,txy,lam,iwk,tvect,trans,perm1,work,tyy,perm0,
     +  x,y,delta,b,sens,aa,bb,type,deltamix,beta,mixturelam,gpflag)  

c	if(dump)then
c          open(unit=203,file='dump.out',form='formatted',
c     +    access='sequential',status='unknown')
c	endif

	do k=1,3

	  if (k .eq. 1) then 
c	    print*,'beginning warmup period of length',jay(1)
	  elseif(k.eq.2)then
c	    print*,'calculating marginal disbn of transformation'
c	    print*,'using sampling period of length',jay(2)
	  else
	    call clamarginal(mixturelam,lam,jay)
c	    print*,'calculating mixture estimates of coeffts.'
c	    print*,'using sampling period of length',jay(3)
	    mix=.true.
	    trans=.false.
	  endif

c main loop
	  do i=1,jay(k)
	
	    if(dump)then
	      sumdelta=0
	      do j=1,n
	        sumdelta=sumdelta+delta(1,j)
	      enddo
c	      write(unit=203,fmt=100) q,sumdelta
c100	      format(i5,2x,i5)
	    endif

            if((dmod(dble(i),10.d0).lt.0.09d0).and.((q.ne.0).and.
     +      (outlier.eqv..false.)))then
c If outlier=F compute a full cholesky every 10 iterations.
              call fullchol(Rone,Rzero,sse1,sse0,q,txx,txy,gpflag,
     +        lam,iwk,tvect,it,p,trans,perm1,work,tyy,indic,sens)
	    elseif((dmod(dble(i),100.d0).lt.0.009d0).and.((q.ne.0)
     +      .and.(outlier)))then
c If outlier=T then recompute txxtxy every 100 iterations.
	      call calctxxtxy(txx,x,b,n,it,p,delta,txy,aa,bb,y,type,
     +        lam,trans)
            endif

c generate p(indic|delta,y)
	    call gindiccint(n,p,seed,txy,txx,tyy,order,
     +      indic,indicprior,q,sse0,sse1,it,Rzero,Rone,perm1,perm0,
     +      sens,random,lam,lamprior,lnjacob,trans,tvect,work,gpflag)

	    if(outlier)then

c generate from p(delta|indic,y)
	    call gdelta2(n,txx,txy,delta,indic,p,q,sens,
     +      lnjacob,lamprior,deltaprior,lam,tyy,perm1,Rone,Rzero,
     +      sse1,sse0,iwk,it,trans,work,b,y,x,kk,deltamix,
     +      mix,seed,tvect,aa,bb,type,gpflag)
	    endif

	    if(trans.and.(k.eq.2))then
c calculate marginal p(lam|y,indic,delta). Do not sample.
              call cmixlam(sse1,lam,mixturelam,lnjacob,lamprior,n,
     +        seed)
	    endif

c if mix calculate mixture estimates beta

	    if(mix)then
	      call cintmixests(perm1,indic,beta,it,p,
     +        Rone,sse1,q,work,tvect,Rzero,n,txy,lam,sens,
     +        betahat)
	    endif

	    if(gdbug1)then
	      do l=1,p
c		print*,'l,indic(l),perm1(l)',l,indic(l),perm1(l)
	      enddo
c	      print*,'lam,tyy(lam),sse1(lam)',lam,tyy(lam),sse1(lam)
c	      print*,'sse0(lam)',sse0(lam)
	    endif

	  enddo

c	  print*,'finished.'
	enddo

c	if(dump)then
c          endfile(unit=203)
c          close(unit=203)
c        endif


c upon exit:	deltamix=mix. est. for delta.
c		lam=mode of mix. est. for lambda.
c		lamprior=mix. est. for lambda. 
c		beta=mix. est. for beta.	

        do i=1,n
          deltamix(i)=deltamix(i)/jay(3)
        enddo
	do i=1,9
	  lamprior(i)=mixturelam(i)/jay(2)
	enddo
        do i=1,p
	  if(kk(i).gt.0.d0)then
            beta(i)=beta(i)/(dble(jay(3))*kk(i))
	  else
	    beta(i)=0.d0
	  endif
        enddo
	if(gdbug2)then
	  do i=1,9
c            print*,'i,lamprior(i)',i,lamprior(i)
          enddo
          do i=1,p
c            print*,'i,beta(i),kk(i)',i,beta(i),kk(i)
          enddo
	endif
	call destandardisex(n,p,kk,x,txx,txy,it)

	return
	end
c________________________________________________________
	subroutine gindiccint(n,p,seed,txy,txx,tyy,order,
     +  indic,indicprior,q,sse0,sse1,it,Rzero,Rone,perm1,perm0,
     +  sens,random,lam,lamprior,lnjacob,trans,tvect,work,gpflag)

c routines called: permute2,gran,solvindlv,dchex

	integer n,order(*),indic(*),q,perm0(*),perm1(*),p
	integer it,lam
	double precision txy(9,*),txx(it,*),seed,tyy(*),lnjacob(*)
	double precision indicprior(*),Rone(it,*),Rzero(it,*)
	double precision sens,sse1(*),sse0(*),lamprior(*)
	double precision tvect(*),work(*)
	logical random,trans,gpflag

c local variables

	integer i,j,lmaxp,k,l,jj,oq
	parameter(lmaxp=450)
	double precision z(2,2),c(lmaxp),s(lmaxp) 
	double precision prob,l1
	double precision h,gran,lnh,small,overflow,dummy
	logical finish,psd,ldbug1,ldbug2
	parameter(small=1.d-13,ldbug1=.false.,ldbug2=.false.)
	parameter(overflow=700.d0)
c overflow is such that exp(overflow)=Inf and exp(-overflow)=0
c on the machine.

	if(random)then
          call permute2(order,p,seed)
	endif

	do k=1,p
	  psd=.false.
	  i=order(k)

	  if (indic(i).eq.0) then

c add row R(.,q+1) to cholesky Rone

	    do j=1,q
	      if(i.le.perm1(j))then
	        Rone(j,q+1)=txx(i,perm1(j))
	      else
	        Rone(j,q+1)=txx(perm1(j),i)
	      endif
	      do l=1,j-1
	 	Rone(j,q+1)=Rone(j,q+1)-Rone(l,q+1)*Rone(l,j)
	      enddo
	      Rone(j,q+1)=Rone(j,q+1)/Rone(j,j)
	    enddo
	    Rone(q+1,q+1)=txx(i,i)
	    do l=1,q
	      Rone(q+1,q+1)=Rone(q+1,q+1)-(Rone(l,q+1)**2)
	    enddo

c extra lines to check matrix psd. If so, then indic(i).eq.0

	    if (Rone(q+1,q+1).lt.small) then
	      psd=.true.
c	      if(gpflag)then
c	        print*,'matrix psd'
c	      endif
	    else
	      Rone(q+1,q+1)=dsqrt(Rone(q+1,q+1))
	    endif

	    if(.not.psd)then
	      perm1(q+1)=i

	      if(trans)then

c calculate sse1 for all 9 values of lambda
	        do l=1,9
c replaced 21/11/95 to remove O(p) calculation with O(q)
c	           do j=1,p
c	             work(j)=txy(l,j)
c	   	   enddo
c                  call solvindlv(Rone,work,tvect,it,q+1,perm1)
		  do j=1,q+1
	            work(j)=txy(l,perm1(j))
	          enddo
	          call solvlvdash(Rone,work,tvect,it,q+1)
                  sse1(l)=tyy(l)
	          do j=1,q+1
c                   sse1(l)=sse1(l)-sens*(tvect(j)**2)/(1.d0+sens)
		    sse1(l)=sse1(l)-tvect(j)**2
	          enddo
	        enddo

	      else
c calculate sse1 for lam only
c replaced 21/11/95 to remove O(p) calculation with O(q)
c	 	do j=1,p
c                  work(j)=txy(lam,j)
c		  if(ldbug1)then
c		    print*,'txy(lam,j),j',txy(lam,j),j
c		  endif
c                enddo
c                call solvindlv(Rone,work,tvect,it,q+1,perm1)

		do j=1,q+1
		  work(j)=txy(lam,perm1(j))
		enddo
		call solvlvdash(Rone,work,tvect,it,q+1)

                sse1(lam)=tyy(lam)
                do j=1,q+1
c                 sse1(lam)=sse1(lam)-sens*(tvect(j)**2)/(1.d0+sens)
		  sse1(lam)=sse1(lam)-(tvect(j)**2)
                enddo
c		if(ldbug1)then
c                  print*,'tyy(lam),lam',tyy(lam),lam
c		  print*,'sse1(lam)',sse1(lam)
c		endif
	      endif

	    endif

	  else 

c remove a colm from the cholesky.
c jj = colm of Rzero which corresponds to i.
c find jj s.t. perm0(jj)=i

	    finish = .false.
	    jj=1
	    do while ((finish.eqv..false.).and.(jj.le.q))
	      if (perm0(jj).eq.i) then
	        finish = .true.
	      else
	        jj=jj+1
	      endif
	    enddo

	    if (jj.lt.q) then

c circular shift colms of R from 1,...,q to 1,..,jj-1,jj+1,..,q,jj

	      call dchex(Rzero,it,q,jj,q,z,2,0,c,s,2)

c reset perm0

	      do l=jj,q-1
	        perm0(l)=perm0(l+1)
	      enddo
	    elseif(jj.eq.q+1)then
c	      print*,'jj.eq.q+1- fatal error'
c	      stop
	    endif

	    if(trans)then
c calculate sse0 for all 9 values of lambda
	      do l=1,9
c replaced 21/11/95 to reduce calc. from O(p) to O(q)
c	        do j=1,p
c	     	  work(j)=txy(l,j)
c	        enddo
c	        call solvindlv(Rzero,work,tvect,it,q-1,perm0)

		do j=1,q-1
		  work(j)=txy(l,perm0(j))
		enddo
		call solvlvdash(Rzero,work,tvect,it,q-1)

	        sse0(l)=tyy(l)
                do j=1,q-1
                  sse0(l)=sse0(l)-(tvect(j)**2)
	        enddo
	      enddo

	    else
c calculate sse0 for lam only
c replaced 21/11/95 to reduce calc. from O(p) to O(q)
c	      do j=1,p
c                work(j)=txy(lam,j)
c                if(ldbug1)then
c                  print*,'txy(lam,j),j',txy(lam,j),j
c                endif
c              enddo
c              call solvindlv(Rzero,work,tvect,it,q-1,perm0)

	      do j=1,q-1
		work(j)=txy(lam,perm0(j))
	      enddo
	      call solvlvdash(Rzero,work,tvect,it,q-1)

              sse0(lam)=tyy(lam)
              do j=1,q-1
c               sse0(lam)=sse0(lam)-sens*(tvect(j)**2)/(1.d0+sens)
		sse0(lam)=sse0(lam)-(tvect(j)**2)
              enddo
c	      if(ldbug1)then
c              print*,'sse0(lam),tyy(lam),lam',sse0(lam),tyy(lam),lam
c	      endif
	    endif

	  endif

	  if (indic(i).eq.0)then
	    oq=q
	  else
	    oq=q-1
	  endif

	  if(trans)then
c calculate the probability indic(i)=1 having integrated over
c lambda.
	    prob=0.d0
	    h=0.d0
            do j=1,9
	     l1=0.d0	
              do l=1,9
	        dummy=lnjacob(l)-lnjacob(j)+dlog(lamprior(l))-
     +          dlog(lamprior(j))+(dble(n)/2.d0)*(dlog(sse0(j))-
     +          dlog(sse1(l)))
c overflow/underflow check for converting back from log scale.
	        if(dummy.gt.overflow)then
	          l1=l1+dexp(overflow)
	 	elseif(dummy.lt.-overflow)then
	 	  l1=l1+dexp(-overflow)
		else
		  l1=l1+dexp(dummy)
	        endif
	      enddo
	      h=h+1.d0/l1
	    enddo
	    h=h*((1.d0-indicprior(i))/indicprior(i))*dsqrt(sens+1.d0)

	  else
c calculate the prob indic(i)=1 for fixed lambda=lam
            lnh=dlog(1.d0-indicprior(i))-dlog(indicprior(i))
            lnh=lnh+dlog(sens+1.d0)/2.d0
            lnh=lnh+(dlog(sse1(lam))-dlog(sse0(lam)))*(dble(n)/2.d0)
	    if(lnh.gt.overflow)then
              h=dexp(overflow)
	    elseif(lnh.lt.-overflow)then
	      h=dexp(-overflow)
	    else
	      h=dexp(lnh)
	    endif
	  endif
	  prob=1.d0/(1.d0+h)

	  if(ldbug2)then
c	    print*,'i,h,prob,indic(i)',i,h,prob,indic(i)
	  endif

c generate indicator variable and reset b's, det's, R's and perm's

          if ((gran(seed).le.prob).and.(.not.psd)) then
	    if(trans)then
	      do l=1,9
	        sse0(l)=sse1(l)
	      enddo
	    else
	      sse0(lam)=sse1(lam)
	    endif
	    if (indic(i) .eq. 0) then
	      q=q+1
	      perm0(q)=perm1(q)
	      do l=1,q
	        Rzero(l,q)=Rone(l,q)
	      enddo
	    else
	      do l=1,q
	        do j=1,l
	          Rzero(j,l)=Rone(j,l)
	        enddo
	        perm0(l)=perm1(l)
	      enddo
	    endif
	    indic(i)=1

          else

	    if(trans)then
	      do l=1,9
	        sse1(l)=sse0(l)
	      enddo
	    else
	      sse1(lam)=sse0(lam)
	    endif
	    if (indic(i) .eq. 1) then
	      q=q-1
	      do l=1,q
                do j=1,l
                  Rone(j,l)=Rzero(j,l)
                enddo
                perm1(l)=perm0(l)
              enddo
	    endif
	    indic(i)=0

c note that if orgindic=0 and R1=R0 no changes are needed

          endif

c go to next indicator variable.

        enddo

	return
	end	
c________________________________________________________
	subroutine cmixlam(sse1,lam,mixturelam,
     +  lnjacob,lamprior,n,seed)

c routines called: none.

	integer lam,n
	double precision sse1(*),mixturelam(*)
	double precision lnjacob(*),lamprior(*),seed
	double precision lnterm,overflow
	parameter(overflow=700.d0)

c local variables

	integer k,j
	double precision prob(9),term

c calculate probabilities

	do j=1,9
	  term=0.d0
	  do k=1,9
	    if(k.ne.j)then
	      lnterm=lnjacob(k)-lnjacob(j)+dlog(lamprior(k))
     +        -dlog(lamprior(j))+(dble(n)/2.d0)*(dlog(sse1(j))-
     +        dlog(sse1(k)))
c altered 28/10/95 for overflow/underflow calculations.
	      if(lnterm.gt.overflow)then
	        term=term+dexp(overflow)
	      elseif(lnterm.lt.-overflow)then
	        term=term+dexp(-overflow)
	      else
	        term=term+dexp(lnterm)
	      endif
	    endif
	  enddo
	  prob(j)=1.d0/(1.d0+term)
	enddo
	
	do j=1,9
	  mixturelam(j)=mixturelam(j)+prob(j)
	enddo

	return
	end
c__________________________________________________________________
	subroutine fullchol(Rone,Rzero,sse1,sse0,q,txx,txy,gpflag,
     +  lam,iwk,tvect,it,p,trans,perm1,work,tyy,indic,sens)

c calculates  Rone, Rzero, sse1, sse0 given txx, perm1, tyy, txy.
c If psd, then indic changed so that it no longer is so. If trans=T
c then all nine values for sse1 and sse0 are calculated.

c routines called: newdchdc, solvindlv

	integer iwk(*),q,lam,it,p,perm1(*),indic(*)
	double precision Rone(it,*),Rzero(it,*),txx(it,*) 
	double precision tvect(*),sse1(*),sse0(*),txy(9,*),tyy(*)
	double precision work(*),sens
	logical trans,gpflag

c local variables

	integer l,j,info
	logical ldbug

	ldbug=.false.

        do j=1,q
          do l=1,j
	    if(perm1(l).le.perm1(j))then
              Rone(l,j)=txx(perm1(l),perm1(j))
	    else
	      Rone(l,j)=txx(perm1(j),perm1(l))
	    endif
          enddo
        enddo

	if(ldbug)then
	  do l=1,p	
c	    print*,'perm1(l),txx(1,l)',perm1(l),txx(1,l)
	  enddo
	endif

        call newdchdc(Rone,it,q,tvect,iwk,0,info)
        if(info.ne.q)then
	  if(gpflag)then
c            print134,'update cholesky readjusting matrix to be psd.',
c     +      'q reduced from',q,'to',info
c134         format(a45,1x,a19,1x,i3,a3,i3)
	  endif
	  do j=info+1,q
	    indic(perm1(j))=0
	  enddo
	  q=info
        endif

	if(trans)then
c find sse for all nine possible transformations.
          do j=1,9
c solve for tvect, Rone'(tvect)=x'y~(j)=work
c replaced 21/11/95 to reduce calc from O(p) to O(q)
c            do l=1,p
c              work(l)=txy(j,l)
c            enddo
c            call solvindlv(Rone,work,tvect,it,q,perm1)

	    do l=1,q
	      work(l)=txy(j,perm1(l))
	    enddo
	    call solvlvdash(Rone,work,tvect,it,q)
		
c calculate sse's
            sse1(j)=tyy(j)
            do l=1,q
c             sse1(j)=sse1(j)-sens*(tvect(l)**2)/(1.d0+sens)
	      sse1(j)=sse1(j)-tvect(l)**2
            enddo
            sse0(j)=sse1(j)
          enddo
	else
c find sse for untransformed data
c replaced 21/11/95 to reduce calc from O(p) to O(q)
c	  do l=1,p
c            work(l)=txy(lam,l)
c          enddo
c          call solvindlv(Rone,work,tvect,it,q,perm1)

	  do l=1,q
            work(l)=txy(lam,perm1(l))
	  enddo
	  call solvlvdash(Rone,work,tvect,it,q)

c calculate sse's
          sse1(lam)=tyy(lam)
          do l=1,q
c           sse1(lam)=sse1(lam)-sens*(tvect(l)**2)/(1.d0+sens)
	    sse1(lam)=sse1(lam)-(tvect(l)**2)
          enddo
          sse0(lam)=sse1(lam)
	endif

c set Rzero=Rone
        do l=1,q
          do j=1,l
            Rzero(j,l)=Rone(j,l)
          enddo
        enddo

	return
	end
c________________________________________________________
	subroutine cintmixests(perm1,indic,beta,it,p,
     +  Rone,sse1,q,work,tvect,Rzero,n,txy,lam,sens,
     +  betahat)

c routines called: calcbetahat

	integer it,p,perm1(*),indic(*),q,n,lam
	double precision beta(*),betahat(*),sens
	double precision sse1(*),Rone(it,*),work(*),tvect(*)
	double precision Rzero(it,*),txy(9,*)

c local variables

	integer i

c calculate beta. Note that we descale this mixture estimate
c for beta at the end of the program.

	call calcbetahat(txy,betahat,lam,work,it,Rone,
     +  q,perm1,p)
c note here that we calculate the mixture estimate of beta under
c the diffuse prior on beta (ie sens=infinity)
c	betahat(i)=betahat(i)*sens/(1.d0+sens)
        do i=1,q
          beta(perm1(i))=beta(perm1(i))+betahat(i)
        enddo

	return
	end
c_______________________________________________________
	subroutine partialupdatexy(txx,txy,tyy,x,y,n,p,delta,it,b,
     +  i,aa,bb,type,lam,trans)

c updates txx,txy,tyy according to a change in a single delta given
c by i

c routines called: ahch

        integer n,p,delta(2,*),it,i,type,lam
        double precision x(n,*),y(*),tyy(*),txy(9,*),txx(it,*)
        double precision b,aa(*),bb(*),mout
	logical trans

c local variables

        integer j,k
	double precision ahch,f1,f2(9)

        if((delta(1,i).eq.0).and.(delta(2,i).eq.1))then
c if an obsn became an outlier then adjust txx, tyy, txy
c accordingly.
	  f1=(1.d0/b)-1.d0
	  if(trans)then
	    do k=1,9
	      f2(k)=(aa(k)+bb(k)*ahch(y(i),type,k))*f1
	    enddo
	  else
	    f2(lam)=(aa(lam)+bb(lam)*ahch(y(i),type,lam))*f1
	  endif
          do k=1,p
c mout introduced 22/11/95. Loop should be unrolled too.
	    mout=x(i,k)*f1
            do j=1,k
              txx(j,k)=txx(j,k)+x(i,j)*mout
            enddo
	    if(trans)then
	      do j=1,9
                txy(j,k)=txy(j,k)+x(i,k)*f2(j)
	      enddo
	    else
	      txy(lam,k)=txy(lam,k)+x(i,k)*f2(lam)
	    endif
          enddo
	  if(trans)then
	    do k=1,9
              tyy(k)=tyy(k)+(f2(k)**2)/f1
	    enddo
	  else
	    tyy(lam)=tyy(lam)+(f2(lam)**2)/f1
	  endif

c if an outlier became 'un'outlier'ed', alter txx,txy and tyy
c accordingly.

        elseif((delta(1,i).eq.1).and.(delta(2,i).eq.0))then
          f1=1.d0-(1.d0/b)
	  if(trans)then
            do k=1,9
              f2(k)=(aa(k)+bb(k)*ahch(y(i),type,k))*f1
            enddo
	  else
	    f2(lam)=(aa(lam)+bb(lam)*ahch(y(i),type,lam))*f1
	  endif
          do k=1,p
c mout introduced 22/11/95. Loop should be unrolled too.
	    mout=x(i,k)*f1
            do j=1,k
              txx(j,k)=txx(j,k)+x(i,j)*mout
            enddo
	    if(trans)then
              do j=1,9
                txy(j,k)=txy(j,k)+x(i,k)*f2(j)
              enddo
	    else
	      txy(lam,k)=txy(lam,k)+x(i,k)*f2(lam)
	    endif
          enddo
	  if(trans)then
            do k=1,9
              tyy(k)=tyy(k)+(f2(k)**2)/f1
            enddo
	  else
	    tyy(lam)=tyy(lam)+(f2(lam)**2)/f1
	  endif
        endif

        return
        end
c_____________________________________________________________
	subroutine updatexy(txx,txy,tyy,x,y,n,p,delta,it,b,aa,
     +  bb,type,lam,trans)

c Recalculate txx, txy, tyy based on any changes from old delta(1,.)
c to new delta(2,.) for all seven values of lambda.

c routines called: partialupdatexy

	integer n,p,delta(2,*),it,type,lam
	double precision x(n,*),y(*),tyy(*),txy(9,*),txx(it,*)
	double precision b,aa(*),bb(*)
	logical trans

c local variables

	integer i

        do i=1,n
	  call partialupdatexy(txx,txy,tyy,x,y,n,p,delta,it,b,
     +    i,aa,bb,type,lam,trans)
	enddo

	return
	end
c___________________________________________________________________
	subroutine calcbetahat(txy,betahat,lam,work,it,Rone,
     +  q,perm1,p)

c routine calculates betahat from the decomposition
c Rone'Rone=txx given txy and given lam. Note this is the betahat for
c the scaled version of x.

c routines called: solvindlv,solvuv

	integer p,q,lam,perm1(*)
	double precision txy(9,*),betahat(*),work(*)
	double precision Rone(it,*)

c local variables

	integer i

        do i=1,p
          betahat(i)=txy(lam,i)
        enddo
        call solvindlv(Rone,betahat,work,it,q,perm1)
        call solvuv(Rone,work,betahat,it,q,1)

	return
	end
c__________________________________________________________________
	subroutine calctxxtxy(txx,x,b,n,it,p,delta,txy,aa,bb,y,type,
     +  lam,trans)

c routines called: ahch

	integer n,it,p,delta(2,*),type,lam
	double precision txx(it,*),b,x(n,*),txy(9,*),aa(*),bb(*)
	double precision y(*)
	logical trans

c local variables

	integer i,j,k,md,mdi
	double precision ahch,mout,mout2

c calculate integers for unrolled loop.
        md=mod(n,8)
        mdi=md+1

	do j=1,p
	  do i=1,j
c mout introduced 20/11/95. Loop unrolled 21/11/95 .
	    mout=0.d0
c	    do k=1,n
c	      if(delta(2,k).eq.1)then
c	        mout=mout+x(k,j)*x(k,i)/b
c	      else
c	        mout=mout+x(k,j)*x(k,i)
c	      endif
c	    enddo
	    do k=1,md
	      if(delta(2,k).eq.1)then
                mout=mout+x(k,j)*x(k,i)/b
              else
                mout=mout+x(k,j)*x(k,i)
              endif
	    enddo
	    do k=mdi,n,8
              if(delta(2,k).eq.1)then
                mout=mout+x(k,j)*x(k,i)/b
              else
                mout=mout+x(k,j)*x(k,i)
              endif
              if(delta(2,k+1).eq.1)then
                mout=mout+x(k+1,j)*x(k+1,i)/b
              else
                mout=mout+x(k+1,j)*x(k+1,i)
              endif
              if(delta(2,k+2).eq.1)then
                mout=mout+x(k+2,j)*x(k+2,i)/b
              else
                mout=mout+x(k+2,j)*x(k+2,i)
              endif
              if(delta(2,k+3).eq.1)then
                mout=mout+x(k+3,j)*x(k+3,i)/b
              else
                mout=mout+x(k+3,j)*x(k+3,i)
              endif
              if(delta(2,k+4).eq.1)then
                mout=mout+x(k+4,j)*x(k+4,i)/b
              else
                mout=mout+x(k+4,j)*x(k+4,i)
              endif
              if(delta(2,k+5).eq.1)then
                mout=mout+x(k+5,j)*x(k+5,i)/b
              else
                mout=mout+x(k+5,j)*x(k+5,i)
              endif
              if(delta(2,k+6).eq.1)then
                mout=mout+x(k+6,j)*x(k+6,i)/b
              else
                mout=mout+x(k+6,j)*x(k+6,i)
              endif
              if(delta(2,k+7).eq.1)then
                mout=mout+x(k+7,j)*x(k+7,i)/b
              else
                mout=mout+x(k+7,j)*x(k+7,i)
              endif
	    enddo
	    txx(i,j)=mout
	  enddo
	  if(trans)then
	    do i=1,9
c mout introduced 20/11/95. inner loop unrolled 21/11/95.
	      mout=0.d0
c	      do k=1,n
c	        if(delta(2,k).eq.1)then
c		   mout=mout+x(k,j)*(aa(i)+bb(i)*
c     +             ahch(y(k),type,i))/b
c                else
c                   mout=mout+x(k,j)*(aa(i)+bb(i)*
c     +             ahch(y(k),type,i))
c	        endif
c	      enddo
              do k=1,md
                if(delta(2,k).eq.1)then
                  mout=mout+x(k,j)*(aa(i)+bb(i)*
     +            ahch(y(k),type,i))/b
                else
                  mout=mout+x(k,j)*(aa(i)+bb(i)*
     +            ahch(y(k),type,i))
                endif
              enddo
	      do k=mdi,n,8
	        if(delta(2,k).eq.1)then
                  mout=mout+x(k,j)*(aa(i)+bb(i)*
     +            ahch(y(k),type,i))/b
                else
                  mout=mout+x(k,j)*(aa(i)+bb(i)*
     +            ahch(y(k),type,i))
                endif
                if(delta(2,k+1).eq.1)then
                  mout=mout+x(k+1,j)*(aa(i)+bb(i)*
     +            ahch(y(k+1),type,i))/b
                else
                  mout=mout+x(k+1,j)*(aa(i)+bb(i)*
     +            ahch(y(k+1),type,i))
                endif
                if(delta(2,k+2).eq.1)then
                  mout=mout+x(k+2,j)*(aa(i)+bb(i)*
     +            ahch(y(k+2),type,i))/b
                else
                  mout=mout+x(k+2,j)*(aa(i)+bb(i)*
     +            ahch(y(k+2),type,i))
                endif
                if(delta(2,k+3).eq.1)then
                  mout=mout+x(k+3,j)*(aa(i)+bb(i)*
     +            ahch(y(k+3),type,i))/b
                else
                  mout=mout+x(k+3,j)*(aa(i)+bb(i)*
     +            ahch(y(k+3),type,i))
                endif
                if(delta(2,k+4).eq.1)then
                  mout=mout+x(k+4,j)*(aa(i)+bb(i)*
     +            ahch(y(k+4),type,i))/b
                else
                  mout=mout+x(k+4,j)*(aa(i)+bb(i)*
     +            ahch(y(k+4),type,i))
                endif
                if(delta(2,k+5).eq.1)then
                  mout=mout+x(k+5,j)*(aa(i)+bb(i)*
     +            ahch(y(k+5),type,i))/b
                else
                  mout=mout+x(k+5,j)*(aa(i)+bb(i)*
     +            ahch(y(k+5),type,i))
                endif
                if(delta(2,k+6).eq.1)then
                  mout=mout+x(k+6,j)*(aa(i)+bb(i)*
     +            ahch(y(k+6),type,i))/b
                else
                  mout=mout+x(k+6,j)*(aa(i)+bb(i)*
     +            ahch(y(k+6),type,i))
                endif
                if(delta(2,k+7).eq.1)then
                  mout=mout+x(k+7,j)*(aa(i)+bb(i)*
     +            ahch(y(k+7),type,i))/b
                else
                  mout=mout+x(k+7,j)*(aa(i)+bb(i)*
     +            ahch(y(k+7),type,i))
                endif
	      enddo
	      txy(i,j)=mout
	    enddo
	  else
	    i=lam
c mout introduced 20/11/95. Loop unrolled 21/11/95.
	    mout=0.d0
c            do k=1,n
c              if(delta(2,k).eq.1)then
c	        mout=mout+x(k,j)*(aa(i)+bb(i)*
c     +          ahch(y(k),type,i))/b
c              else
c	        mout=mout+x(k,j)*(aa(i)+bb(i)*
c     +          ahch(y(k),type,i))
c              endif
c            enddo
            do k=1,md
              if(delta(2,k).eq.1)then
                mout=mout+x(k,j)*(aa(i)+bb(i)*
     +          ahch(y(k),type,i))/b
              else
                mout=mout+x(k,j)*(aa(i)+bb(i)*
     +          ahch(y(k),type,i))
              endif
            enddo
            do k=mdi,n,8
              if(delta(2,k).eq.1)then
                mout=mout+x(k,j)*(aa(i)+bb(i)*
     +          ahch(y(k),type,i))/b
              else
                mout=mout+x(k,j)*(aa(i)+bb(i)*
     +          ahch(y(k),type,i))
              endif
              if(delta(2,k+1).eq.1)then
                mout=mout+x(k+1,j)*(aa(i)+bb(i)*
     +          ahch(y(k+1),type,i))/b
              else
                mout=mout+x(k+1,j)*(aa(i)+bb(i)*
     +          ahch(y(k+1),type,i))
              endif
              if(delta(2,k+2).eq.1)then
                mout=mout+x(k+2,j)*(aa(i)+bb(i)*
     +          ahch(y(k+2),type,i))/b
              else
                mout=mout+x(k+2,j)*(aa(i)+bb(i)*
     +          ahch(y(k+2),type,i))
              endif
              if(delta(2,k+3).eq.1)then
                mout=mout+x(k+3,j)*(aa(i)+bb(i)*
     +          ahch(y(k+3),type,i))/b
              else
                mout=mout+x(k+3,j)*(aa(i)+bb(i)*
     +          ahch(y(k+3),type,i))
              endif
              if(delta(2,k+4).eq.1)then
                mout=mout+x(k+4,j)*(aa(i)+bb(i)*
     +          ahch(y(k+4),type,i))/b
              else
                mout=mout+x(k+4,j)*(aa(i)+bb(i)*
     +          ahch(y(k+4),type,i))
              endif
              if(delta(2,k+5).eq.1)then
                mout=mout+x(k+5,j)*(aa(i)+bb(i)*
     +          ahch(y(k+5),type,i))/b
              else
                mout=mout+x(k+5,j)*(aa(i)+bb(i)*
     +          ahch(y(k+5),type,i))
              endif
              if(delta(2,k+6).eq.1)then
                mout=mout+x(k+6,j)*(aa(i)+bb(i)*
     +          ahch(y(k+6),type,i))/b
              else
                mout=mout+x(k+6,j)*(aa(i)+bb(i)*
     +          ahch(y(k+6),type,i))
              endif
              if(delta(2,k+7).eq.1)then
                mout=mout+x(k+7,j)*(aa(i)+bb(i)*
     +          ahch(y(k+7),type,i))/b
              else
                mout=mout+x(k+7,j)*(aa(i)+bb(i)*
     +          ahch(y(k+7),type,i))
              endif
	    enddo
	    txy(i,j)=mout
	  endif
	enddo

	return
	end
c________________________________________________________
	subroutine clamarginal(mixturelam,lam,jay)

c 1) finds the maximiser of the marginal density. 2) fixes lam
c to this value.

c routines called: none

	integer lam,jay(*)
	double precision mixturelam(*)

c local variables

	integer i
	double precision maximum

	maximum=0.d0	
	do i=1,9
	  if(mixturelam(i).gt.maximum)then
	     lam=i
	     maximum=mixturelam(i)
	  endif
	enddo

	return
	end
c________________________________________________________
        subroutine standardisex(n,p,kk,x,it)

c routines called: none 

        integer p,n,it
        double precision x(n,*),kk(*)

c normalise x as x=xK^-1 by a diagonal
c matrix K=diag(kk1,...,kkp) where kkj=sqrt(sumoveri(x(i,j)^2))

c local variables

        integer i,j

        do j=1,p
          kk(j)=0.d0
          do i=1,n
            kk(j)=kk(j)+x(i,j)**2
          enddo
          kk(j)=dsqrt(kk(j))
        enddo

        do j=1,p
	  if(kk(j).gt.0.d0)then
	    do i=1,n
              x(i,j)=x(i,j)/kk(j)
            enddo	
	  endif
        enddo

        return
        end
c__________________________________________________________________
        subroutine destandardisex(n,p,kk,x,txx,txy,it)

c routines called: none

        integer p,n,it
        double precision txx(it,*),x(n,*),txy(9,*),kk(*)

c local variables
        integer i,j

        do i=1,p
          do j=1,9
            txy(j,i)=txy(j,i)*kk(i)
          enddo
          do j=1,i
            txx(j,i)=txx(j,i)*kk(i)*kk(j)
          enddo
        enddo

	do j=1,p
          do i=1,n
            x(i,j)=x(i,j)*kk(j)
          enddo
        enddo

        return
        end
c_____________________________________________________________
	subroutine initialise(n,p,q,mix,indic,it,txx,Rone,Rzero,
     +  sse0,sse1,txy,lam,iwk,tvect,trans,perm1,work,tyy,perm0,
     +  x,y,delta,b,sens,aa,bb,type,deltamix,beta,mixturelam,gpflag)

c routines called: ahch, calctxxtxy, fullchol 

	integer n,p,q,indic(*),it,lam,iwk(*),perm1(*),perm0(*)
	integer delta(2,*),type
	double precision txx(it,*),Rone(it,*),Rzero(it,*),aa(*),bb(*)
	double precision sse0(*),sse1(*),txy(9,*),tvect(*),deltamix(*)
	double precision work(*),tyy(*),x(n,*),y(*),b,sens,beta(*)
	double precision mixturelam(*)
	logical mix,trans,gpflag

c local variables
	integer i,j
	double precision ahch

        mix=.false.

c set all deltas to zero (ie: start with no outliers.)

        do i=1,n
          delta(1,i)=0
          delta(2,i)=0
        enddo

c set q, perm1 and perm0
	q=0
	do i=1,p
	  if(indic(i).eq.1)then
	    q=q+1
	    perm1(q)=i
	    perm0(q)=i
	  endif
	enddo

c calculate initial txx,txy,tyy
	call calctxxtxy(txx,x,b,n,it,p,delta,txy,aa,bb,y,type,
     +  lam,trans)
	if(trans)then
	  do j=1,9
	    tyy(j)=0.d0
	    do i=1,n	
	      tyy(j)=tyy(j)+(aa(j)+bb(j)*ahch(y(i),type,j))**2
	    enddo
	  enddo
	else
	  j=lam
	  tyy(j)=0.d0
          do i=1,n    
            tyy(j)=tyy(j)+(aa(j)+bb(j)*ahch(y(i),type,j))**2
          enddo
	endif

	call fullchol(Rone,Rzero,sse1,sse0,q,txx,txy,gpflag,
     +  lam,iwk,tvect,it,p,trans,perm1,work,tyy,indic,sens)

c fix mixture estimates to zero

	do i=1,n
	  deltamix(i)=0.d0
	enddo
	do i=1,p
	  beta(i)=0.d0
	enddo
	do i=1,9
	  mixturelam(i)=0.d0
	enddo

	return
	end
c_______________________________________________________
	subroutine gdelta2(n,txx,txy,delta,indic,p,q,sens,
     +  lnjacob,lamprior,deltaprior,lam,tyy,perm1,Rone,Rzero,
     +  sse1,sse0,iwk,it,trans,work,b,y,x,kk,deltamix,
     +  mix,seed,tvect,aa,bb,type,gpflag)

c generates delta(i)| delta(j.ne.i), y, indic
c delta(i)=1 then obsn i is considered an outlier, otherwise 
c delta(i)=0 and it is not.

c Note (2/96) the Rzero matrix is used as work space
c (Rone is actually used as itself) and Rzero
c is updated so that Rzero=Rone at the end of routine. sse1 and sse0
c are used though.

c routines called: ahch, gran, newdchdc, solvindlv, partialupdatexy

	integer indic(*),n,q,p,delta(2,*),lam,perm1(*)
	integer iwk(*),it,type
	double precision txx(it,*),txy(9,*),tyy(*)
	double precision lnjacob(*),lamprior(*),deltaprior(*)
	double precision sse1(*),sse0(*),Rone(it,*),x(n,*)
	double precision Rzero(it,*),work(*),y(*),b,kk(*)
	double precision deltamix(*),seed,tvect(*),aa(*)
	double precision sens,bb(*)
	logical trans,mix,gpflag

c local variable

	integer i,j,k,l,info
	double precision h,sum,lnsum,g(9),ahch,lnh,mmin,lwkj(9)
	double precision gran,prob,out,in,overflow,mout,lwkk(9)
	double precision xvect(450),zvect(450),sc1,sc2,sc3
c remove with global once routine tested.
	logical ldbug2,ldbug1,psd
	parameter(ldbug2=.false.,ldbug1=.false.,overflow=700.d0)

c set out and in
	out=1.d0-(1.d0/b)
	in=(1.d0/b)-1.d0

	do i=1,n

c set delta(1.) to the old value of delta and generate delta(2.)
	  delta(1,i)=delta(2,i)

c calculate adjustment to y.
	  if(trans)then
            do j=1,9
	      g(j)=aa(j)+bb(j)*ahch(y(i),type,j)
	    enddo
	  else
	    g(lam)=aa(lam)+bb(lam)*ahch(y(i),type,lam)
	  endif

c place x(i,.) into xvect(i,.)
	  do j=1,q
	    xvect(j)=x(i,perm1(j))
	  enddo

c solve for zvect, xvect=Rone'zvect          O(q^2)
	  call solvlvdash(Rone,xvect,zvect,it,q)
	  sc1=0.d0
	  do j=1,q
	    sc1=sc1+zvect(j)**2
	  enddo

c PART1: calculate sse0 if delta(1,i)=1
	  
	  if(delta(1,i).eq.1)then
	    if(trans)then
	      do j=1,9
                mout=g(j)*out

c set work=x'y~(j)+adjustment(i,j)		O(q)
                do l=1,q
                  work(l)=txy(j,perm1(l))+xvect(l)*mout
                enddo

c solve for tvect, work=Rone'tvect		O(q^2)
                call solvlvdash(Rone,work,tvect,it,q)

c calculate sse0
	        sc2=0.d0
	        sc3=0.d0
		do l=1,q
		  sc2=sc2+zvect(l)*tvect(l)
		  sc3=sc3+tvect(l)**2
	        enddo
	        sc2=sc2**2
		sse0(j)=tyy(j)+out*(g(j)**2)
		sse0(j)=sse0(j)-sc3+out*sc2/(1.d0+out*sc1)

              enddo

            else
c find sse for untransformed data
	      mout=out*g(lam)

c set work=x'y~(j)+adjustment(i,j)		O(q)
	      do l=1,q
	        work(l)=txy(lam,perm1(l))+mout*xvect(l)
	      enddo

c solve for tvect, work=Rone'tvect              O(q^2)
              call solvlvdash(Rone,work,tvect,it,q)

c calculate sse0 				O(q)
              sc2=0.d0
              sc3=0.d0
              do l=1,q
                sc2=sc2+zvect(l)*tvect(l)
                sc3=sc3+tvect(l)**2
              enddo
              sc2=sc2**2
              sse0(lam)=tyy(lam)+out*(g(lam)**2)
              sse0(lam)=sse0(lam)-sc3+out*sc2/(1.d0+out*sc1)

	    endif

c PART2: Calculate sse1 for case where delta(1,i)=0
          else

            if(trans)then
              do j=1,9
	        mmin=g(j)*in

c set work=x'y~(j)+adjustment(i,j)              O(q)
                do l=1,q
                  work(l)=txy(j,perm1(l))+xvect(l)*mmin
                enddo

c solve for tvect, work=Rone'tvect              O(q^2)
                call solvlvdash(Rone,work,tvect,it,q)

c calculate sse1
                sc2=0.d0
                sc3=0.d0
                do l=1,q
                  sc2=sc2+zvect(l)*tvect(l)
                  sc3=sc3+tvect(l)**2
                enddo
                sc2=sc2**2
                sse1(j)=tyy(j)+in*(g(j)**2)
                sse1(j)=sse1(j)-sc3+in*sc2/(1.d0+in*sc1)

              enddo
            else

              mmin=in*g(lam)

c set work=x'y~(j)+adjustment(i,j)              O(q)
              do l=1,q
                work(l)=txy(lam,perm1(l))+mmin*xvect(l)
              enddo

c solve for tvect, work=Rone'tvect              O(q^2)
              call solvlvdash(Rone,work,tvect,it,q)

c calculate sse1
              sc2=0.d0
              sc3=0.d0
              do l=1,q
                sc2=sc2+zvect(l)*tvect(l)
                sc3=sc3+tvect(l)**2
              enddo
              sc2=sc2**2
              sse1(lam)=tyy(lam)+in*(g(lam)**2)
              sse1(lam)=sse1(lam)-sc3+in*sc2/(1.d0+in*sc1)

            endif

          endif

c PART3: calculate prob(delta(2,i)=1)
	  if(trans)then

c bug in loop removed 11/2/96
	    h=0.d0	
c local work vectors introduced 21/11/95 to remove repeated calculations.
	    mout=dble(n)/2.d0
	    do j=1,9
	      lwkj(j)=-lnjacob(j)-dlog(lamprior(j))+mout*dlog(sse0(j))
	      lwkk(j)=lnjacob(j)+dlog(lamprior(j))-mout*dlog(sse1(j))
	    enddo
	    do j=1,9
	      sum=0.d0
	      do k=1,9
		lnsum=lwkk(k)+lwkj(j)
c adjusted 28/10/95 for overflow/underflow calculations.
	        if(lnsum.gt.overflow)then
		  sum=sum+dexp(overflow)
		elseif(lnsum.lt.-overflow)then
	          sum=sum+dexp(-overflow)
		else
	          sum=sum+dexp(lnsum)
		endif
	      enddo
	      h=h+1.d0/sum
	    enddo

	  else
	    lnh=(dble(n)/2.d0)*(dlog(sse1(lam))-dlog(sse0(lam)))
	    if(lnh.gt.overflow)then
	      h=dexp(overflow)
	    elseif(lnh.lt.-overflow)then
	      h=dexp(-overflow)
	    else
	      h=dexp(lnh)
	    endif
	  endif	    
	  
	  h=h*dsqrt(b)*((1.d0-deltaprior(i))/deltaprior(i))
	  prob=1.d0/(1.d0+h)

	  sc1=gran(seed)

c four cases:-
c 	(a) delta(1,i)=0 and sc1 > prob (no change)
c	(b) delta(1,i)=1 and sc1 < prob (no change)
c	(c) delta(1,i)=0 and sc1 < prob (change but with check)
c	(d) delta(1,i)=1 and sc1 > prob (change no check)

c case (a)
	  if((sc1.ge.prob).and.(delta(1,i).eq.0))then
	    delta(2,i)=0
	    if(trans)then
	      do j=1,9
	        sse1(j)=sse0(j)
	      enddo
	    else
              sse1(lam)=sse0(lam)
            endif

c case (b)
	  elseif((sc1.lt.prob).and.(delta(1,i).eq.1))then
            delta(2,i)=1
            if(trans)then
              do j=1,9
                sse0(j)=sse1(j)
              enddo
            else
              sse0(lam)=sse1(lam)
            endif

c case (c)
	  elseif((sc1.lt.prob).and.(delta(1,i).eq.0))then
c update cholesky with `downdating'	O(q^2)+O(q)
	    do j=1,q
	      work(j)=xvect(j)*dsqrt(out)
	    enddo
	    call newdchdd(Rone,it,q,work,Rzero,it,0,1.d0,1.d0,zvect,
     +      tvect,info)
	    if(info.eq.-1)then
	      psd=.true.
	    else
	      psd=.false.
	    endif
c if psd make no change in delta and set sse1=sse0
	    if(psd)then
	      prob=0.d0
	      delta(2,i)=0
              if(trans)then
                do j=1,9
                  sse1(j)=sse0(j)
                enddo
              else
                sse1(lam)=sse0(lam)
              endif
	    else
c set sse0=sse1 and update txx, txy etc.
	      delta(2,i)=1
              if(trans)then
                do j=1,9
                  sse0(j)=sse1(j)
	        enddo
              else
                sse0(lam)=sse1(lam)
              endif
              call partialupdatexy(txx,txy,tyy,x,y,n,p,delta,it,b,i,
     +        aa,bb,type,lam,trans)
	    endif

c case (d)
	  else
	    delta(2,i)=0
            do j=1,q
              work(j)=xvect(j)*dsqrt(out)
            enddo
	    call newdchud(Rone,it,q,work,Rzero,it,0,1.d0,1.d0,
     +      zvect,tvect)
	    if(trans)then
              do j=1,9
                sse1(j)=sse0(j)
              enddo
            else
              sse1(lam)=sse0(lam)
            endif
            call partialupdatexy(txx,txy,tyy,x,y,n,p,delta,it,b,i,
     +      aa,bb,type,lam,trans)
          endif

	  if(mix)then
	    deltamix(i)=deltamix(i)+prob
	  endif

c go around for next outlier variable.
	enddo

c update Rzero

        do k=1,q
          do j=1,k
            Rzero(j,k)=Rone(j,k)
          enddo
        enddo

	return
	end
c_____________________________________________________________
c fortran header for fitting splines by bayesian model selection in an
c Includes two types of transformations (indexed by type) 9 levels
c in each. The tranformation is of the form y |--> aa + bb * ahch(y)

	subroutine breg4_2(n,p1,p2,xms,xspline,y,density,xplot,yplot,
     +  iwkn,indic,jay,sens,plotlength,numk,knot,
     +  sxs,txy,txx,it,resids,indicprior,seed,fpres,lam,
     +  lamprior,trans,z,beta,x,outlier,delta,deltaprior,deltamix,
     +  Rone,Rzero,type,wk1,wk2,aa,bb,random,b)

c routines called: ahch, plusop, dpvbubsort, constructx, calcjacob
c constructz, allocvarnames, bms4_2, cmixf,

	integer n,p1,p2,density(*),numk(*),indic(*),jay(*)
	integer plotlength,it,delta(2,*),iwkn(*),lam,type
	double precision xms(n,*),xspline(n,*),y(*),deltaprior(*)
	double precision xplot(plotlength,*),yplot(plotlength,*)
	double precision knot(p2,*),deltamix(*)
	double precision sens,sxs(*),resids(*),txx(it,*),txy(9,*)
	double precision lamprior(*),indicprior(*),seed,b
	double precision fpres(n,*),z(plotlength,p2,*),Rzero(it,*)
	double precision beta(*),aa(*),bb(*),x(n,*),Rone(it,*)
	double precision wk1(n,*),wk2(*)
	logical trans,outlier,random,dbug1
	parameter(dbug1=.false.)

c xplot the x grid for the mixture and modal estimates.
c yplot the corresponding modal estimates of y for xplot
c beta: beta for mixture estimate.
c resids(*): residuals from fit.

c local variables

	integer gmaxp,p
        parameter(gmaxp=450)
	integer g,h,i,j,k,l,jj
	integer vble(gmaxp),order(gmaxp)
	double precision tyy(9),lnjacob(9)
	double precision ahch,scale
	logical keepgoing, intercept

c	print*,'Nonparametric Bayesian Regression: v4.1'	
	intercept=.true.
	p=0
c	print*,'constructing matrices'
	if (intercept) then
	  p=1
	endif
	p=p+p1+3*p2

c Construct knots

	k=0
	do l=1,p2

c sort each colm of xspline into ascending order

	  do g=1,n
	    sxs(g)=xspline(g,l)
	  enddo

c sort and calculate scale

	  call getpos(sxs,n,iwkn)
	  call sort1(sxs,wk2,n,iwkn)
	  do g=1,n
	    sxs(g)=wk2(g)
	  enddo
	  scale=sxs(n)-sxs(1)

c as a sideline, can construct xplot(.,l)

	  do h=1,plotlength
	    xplot(h,l)=sxs(1)+(dble(h-1)*scale)/dble(plotlength)
	  enddo
	  keepgoing=.true.
	  j=0
	  jj=0
	  oldknot=sxs(1)
	  do while(keepgoing)
	    j=j+1
	    jj=jj+1
	    knot(l,jj)=(sxs(j*density(l)+1)+sxs(j*density(l)))/2.d0

c If knot is not equal to the old knot then include. This is
c overcome discrete observations in the data.

	    if ((knot(l,jj)-oldknot).ge.(scale*1.d-6))then
	      oldknot=knot(l,jj)
	    else
	      jj=jj-1
	    endif

	    if((j+2)*density(l).gt.n)then
	      keepgoing=.false.
	      numk(l)=jj
	      p=p+jj
	    endif
	  enddo
	 
	enddo 

	do i=1,p
	  order(i)=i
	enddo
	
	call constructx(x,xms,xspline,knot,numk,n,p1,p2,
     +  p,intercept)

c construct lnjacob for the transformation aa + bb * ahch(.)

	call calcjacob(lnjacob,n,y,type,wk1,wk2,aa,bb,trans,lam)
	if(dbug1)then
	  do i=1,9
c	    print*,'aa(i),bb(i),i',aa(i),bb(i),i
	  enddo
	endif

c construct z, where z  holds the vector of spline terms for each
c of the p2 spline variables at each point on plotlength.

	call constructz(numk,knot,z,plotlength,p2,xplot)

c allocate variable names to each indicator variable.

 	call allocvarnames(vble,p1,p2,numk,intercept)	

c having constructed matricies, perform model selection.

	call bms4_2(n,p,txx,txy,tyy,indic,jay,
     +  it,Rone,Rzero,order,sens,indicprior,random,seed,lam,
     +  lnjacob,lamprior,trans,beta,vble,x,y,b,delta,
     +  outlier,deltaprior,deltamix,aa,bb,type)

c calculate mixture estimates and standard deviations.

	call cmixf(beta,it,z,plotlength,p2,
     +  yplot,p1,vble,p)

c calculate resids, fpres (on transformed scale.)

	do i=1,n
	  resids(i)=ahch(y(i),type,lam)*bb(lam)+aa(lam)
	enddo  

	do j=1,p
	  do i=1,n
	    resids(i)=resids(i)-x(i,j)*beta(j)
	  enddo
	enddo

	if(intercept)then
	  k=1
	else
	  k=0
	endif
	k=k+p1

	if(p2.ge.1)then
	  do j=1,p2
	    do i=1,n
	      fpres(i,j)=resids(i)
	    enddo
	  enddo

	  do j=1,p2
	    do l=1,(3+numk(j))
	      k=k+1
	      do i=1,n
	        fpres(i,j)=fpres(i,j)+beta(k)*x(i,k)
	      enddo
	    enddo
	  enddo
	     
	endif

	return
	end	
c_______________________________________________________________
	subroutine constructz(numk,knot,z,plotlength,p2,xplot)

c routines called: plusop

	integer p2,numk(*),plotlength
	double precision knot(p2,*),z(plotlength,p2,*)
	double precision xplot(plotlength,*)

c local variables

	integer i,j,k
	double precision plusop

c loop around for each splined variable
	do i=1,p2

c loop around for each point of xplot
	  do j=1,plotlength

	    z(j,i,1)=xplot(j,i)
	    z(j,i,2)=xplot(j,i)**2
            z(j,i,3)=xplot(j,i)**3
	    do k=1,numk(i)
	      z(j,i,k+3)=plusop(xplot(j,i)-knot(i,k))**3
	    enddo

	  enddo
	enddo

	return
	end
c____________________________________________________________
	subroutine allocvarnames(vble,p1,p2,numk,intercept)

c routine allocates the variable number from which each indicator
c came from.

c routines called: none

	integer p1,p2,vble(*),numk(*)
	logical intercept

c local variables

	integer i,j,k

	k=1
	if(intercept)then
	  vble(k)=0
	  k=k+1
	endif

c for each of the colms associated with a linear term.
	do i=1,p1
	  vble(k)=i
	  k=k+1
	enddo

c for each of the spline variables.
	do i=1,p2
	  do j=1,(3+numk(i))
	    vble(k)=i+p1
	    k=k+1
	  enddo
	enddo

	return
	end
c___________________________________________________________
	subroutine cmixf(beta,it,z,plotlength,p2,
     +  yplot,p1,vble,p)

c routines called: none

	integer p2,it,plotlength,p1,vble(*),p
	double precision beta(*)
	double precision z(plotlength,p2,*),yplot(plotlength,*)

c local variables

	integer i,j,k

c first set the plots to zero.

	do i=1,plotlength
	  do k=1,p2
	    yplot(i,k)=0.d0
	  enddo
	enddo

	do i=1,plotlength
	  j=1
	  do k=1,p
c go through each variable and if it is a splined variable, add its
c contribution onto the mixture estimate for that variable.
	    if((vble(k)-p1).gt.0)then
	      yplot(i,vble(k)-p1)=yplot(i,vble(k)-p1)+z(i,
     +        vble(k)-p1,j)*beta(k)
	    endif
            if(vble(k+1).ne.vble(k))then
              j=1
            else
              j=j+1
            endif
	  enddo
	enddo

	return
	end
c_______________________________________________________________
	subroutine constructx(x,xms,xspline,knot,numk,n,p1,p2,
     +  p,intercept)

c routines called: plusop

	integer p,n,p1,p2,numk(*)
	double precision x(n,*),xms(n,*),xspline(n,*)
	double precision knot(p2,*)
	logical intercept

c local variables

	integer i,j,k,l
	double precision plusop

	k=0
	if(intercept)then
	  k=k+1
	  do i=1,n
	    x(i,k)=1.d0
	  enddo
	endif
	do j=1,p1
	  k=k+1
	  do i=1,n
	    x(i,k)=xms(i,j)
	  enddo
	enddo
	do j=1,p2
	  do i=1,n
	    x(i,k+1)=xspline(i,j)
	    x(i,k+2)=xspline(i,j)**2
	    x(i,k+3)=xspline(i,j)**3
	  enddo
	  k=k+3
	  do l=1,numk(j)
	    k=k+1
	    do i=1,n
	      x(i,k)=plusop(xspline(i,j)-knot(j,l))**3
	    enddo
	  enddo
	enddo
	
	if(k.ne.p)then
c	  print*,'k.ne.p in construct x'
c	  stop
	endif

	return
	end
c_______________________________________________________
c___________________________________________________
c function to sort a dp vector x into order (largest at 1
c smallest at n)
        subroutine dpvbubsort(x,n)

        integer n,i,j
        double precision x(*),temp
	logical finished

	finished=.false.
	i=2
	do while((.not.finished).and.(i.le.n))
	  finished=.true.
	  do j=n,i,-1
	    if(x(j).gt.x(j-1))then
	      temp=x(j-1)
	      x(j-1)=x(j)
	      x(j)=temp
	      finished=.false.
	    endif
	  enddo
	  i=i+1
	enddo
	
        return
        end
c__________________________________________________________
        double precision function plusop(x)

c calculates the max(x,0.d0)

        double precision x

        if(x .gt. 0.d0) then
          plusop=x
        else
          plusop=0.d0
        endif

        return
        end
c_____________________________________________________________
        subroutine permute2(perm,p,seed)

c returns a random permutation of the p integers stored in
c perm. Maximum possible permutation size is maxsize.

        integer perm(*),p
        double precision seed

c local variables

        integer maxsize
        parameter(maxsize=200)
        integer i,iunif,tempperm(maxsize),in

        do i=1,p
          in=iunif(seed,p+1-i)
          tempperm(i)=perm(in)
          perm(in)=perm(p+1-i)
        enddo
        do i=1,p
          perm(i)=tempperm(i)
        enddo

        return
        end
c__________________________________________________________
        integer function iunif(seed,n)

c generates an integer from a uniform distribution over 1,..,n

        integer n
        double precision seed

c local variables

        double precision gran

        iunif=int(gran(seed)*n)+1

        return
        end
c_____________________________________________________
c solve for y,Ly = x where L is a lower triangular matrix and
c x and y are vectors. L is stored in upper triangle as Ldash
      subroutine solvlvdash(ldash,x,y,il,m)
      integer il,m,i,j
      double precision ldash,x,y,sum,zero
      dimension ldash(il,1),x(1),y(1)
      data zero/0.d0/
      do 100 i = 1,m
        sum = zero
        if(i .gt. 1)then
           do 50  j = 1,i-1
              sum = sum + ldash(j,i)*y(j)
 50        continue
        endif
        y(i) = (x(i) - sum)/ldash(i,i)
 100  continue
      return
      end

c solve for y,Ly = x where L is a lower triangular matrix and
c x and y are vectors. L is stored in upper triangle as Ldash.
c The x vector is stored in x with indicies ind(1),...,ind(m)
c 
      subroutine solvindlv(ldash,x,y,ildash,m,ind)

c routines called: none.

      integer ildash,m,i,j,ind(*)
      double precision ldash(ildash,*),x(*),y(*)
      double precision sum,zero
      data zero/0.d0/
      do 100 i = 1,m
        sum = zero
        if(i .gt. 1)then
           do 50  j = 1,i-1
              sum = sum + ldash(j,i)*y(j)
 50        continue
        endif
        y(i) = (x(ind(i)) - sum)/ldash(i,i)
 100  continue
      return
      end

c solve for y , uy = x, where u is upper triangular m x m matrix
c and y and x are m x 1 vectors. 
c istore is an indicator variable. If istore = 1, then u is stored in upper
c triangular form, and if istore = 0, then u is stored in lower triangular
c form.  
      subroutine solvuv(u,x,y,iu,m,istore)

c routines called: none.

      integer iu,m,istore,i,j
      double precision u,x,y,sum,zero
      dimension u(iu,1),x(1),y(1)
      data zero/0.d0/
      do 100 j = m,1,-1
         sum = zero
         if(j .lt. m)then
            do 50 i=j+1,m
                if(istore .eq. 1)then
                    sum = sum +u(j,i)*y(i)
                else
                    sum = sum + u(i,j)*y(i)
                endif 
50           continue
          endif
             y(j) = (x(j) - sum)/u(j,j)
100   continue
      return
      end

c______________________________________________________

c
c	Subroutine to generate a normal density
c
	double precision function fnorml(x,mean,stdev)
c
c	     x  double precision               input  Density function argument
c	  mean  double precision               input  Mean
c	 stdev  double precision               input  Standard deviation
c
c	Subroutines called:  None
c
c	Precision:  Double
c
c	Arguments:
c
	double precision x, mean, stdev
c
c	Local variables:
c
	double precision isr2pi, two
	data isr2pi / .398942280401433d0 /
	data two / 2.d0 /
	isr2pi=  .398942280401433d0 
	two = 2.d0 

	
c
	fnorml = isr2pi * dexp(-(x-mean)**2 / (two*stdev**2)) / stdev
	return
	end
c__________________________________________________________________________
c
c       Function to generate a single U(0,1) deviate
c
        double precision function gran(dseed)
c
c        dseed  double precision     input/  Seed for newranu
c                                    output
c
c       Subroutines called:  newranu
c
c       Precision:  Double
c
c       Arguments:
c
        double precision dseed
c
c       Local variables:
c
        double precision u
        call newranu(dseed,1,u)
        gran = u
        return
        end
c__________________________________________________________________________

c This routine calculates the log of the determinant of the
c absolute value of a jacobian. ie lnjacob(i)=log(|det(J(i))|)
c where J(i) corresponds to the jacobian for transformation indexed
c by integer i=1,..,9

c The type of transformation considered here is of the IQR and Med
c invariant form where z_j=a(y)+b(y)h(y_j) and h(.) is either of
c 1) the box-cox form or 2) of the [0,1]->[-inf,inf] form.

c For type 1) we consider -y^-2, -y^-1.5, -y^-1, -y^-0.5, log(y),
c y^0.5, y^1, y^1.5, y^2
c For type 2) we consider Psi^-1(y^0.1), Psi^-1(y^0.25), Psi^-1(y^0.5),
c Psi^-1(y^0.75), Psi^-1(y), Psi^-1(y^1.5), Psi^-1(y^2), logit(y),
c log(-log(1-y)).

c note that wk2(,1-3) contains the three non-zero off diag colms
c and wk2(,4) is the diagonal.

	subroutine calcjacob(lnjacob,n,y,type,wk1,wk2,aa,bb,trans,
     +  lam)

c routines called: ahch, getpos, pdh

	integer n,type,wk2(*),lam
	double precision lnjacob(*),y(*),wk1(n,*),aa(*),bb(*)
	logical trans

c local variables
	integer i,j
	integer q1,q2,q3
	double precision pdh,ahch
	double precision medy,medhy,iqry,iqrhy

c get sorted permutation and perturb (if nec.)

	call getpos(y,n,wk2)
c note: the following code is for perturbing the data. 1)It is not
c correct 2)the issue needs thinking through with regards to how
c we treat discrete data.
c	small=(y(wk2(n))-y(wk2(1)))/(dble(n)*1000000.d0)
c	do j=1,n
c	  if(y(wk2(j)).eq.y(wk2(j+1)))then
c	    y(wk2(j+1))=y(wk2(j+1))+small
c	  endif
c	enddo

c calculate constants.
	q1=n/4
	q2=n/2
	q3=3*n/4

	medy=y(wk2(q2))
	iqry=y(wk2(q3))-y(wk2(q1))

c loop around for all the different nine transformations.

	if(trans)then

	  do i=1,9
	    lnjacob(i)=0.d0
	    medhy=ahch(medy,type,i)
	    iqrhy=ahch(y(wk2(q3)),type,i)-ahch(y(wk2(q1)),type,i)
            bb(i)=iqry/iqrhy
            aa(i)=medy-bb(i)*medhy

c fill in non-diagonal elements of the (ordered data) jacobian.
	    do j=q1+1,n
	      wk1(j,1)=(pdh(y,i,wk2(j),type)*iqry-iqrhy)*
     +        (ahch(y(wk2(j)),type,i)-medhy)/(iqrhy**(2.d0))
	      wk1(j,2)=1.d0-pdh(y,i,wk2(j),type)*bb(i)
	      wk1(j,3)=(iqrhy-pdh(y,i,wk2(j),type)*iqry)*
     +        (ahch(y(wk2(j)),type,i)-medhy)/(iqrhy**(2.d0))
	    enddo

c fill in 3 key diagonal elements
	    wk1(q1,1)=(pdh(y,i,wk2(q1),type)*iqry-iqrhy)*
     +      (ahch(y(wk2(q1)),type,i)-medhy)/(iqrhy**(2.d0))+
     +      pdh(y,i,wk2(q1),type)*bb(i)
	    wk1(q2,2)=1.d0
	    wk1(q3,3)=(iqrhy-pdh(y,i,wk2(q3),type))*
     +      (ahch(y(wk2(q3)),type,i)-medhy)/(iqrhy**(2.d0))+
     +      pdh(y,i,wk2(q3),type)*bb(i)

c fill in remaining diagonal elements
	    do j=1,n
	      if((j.ne.q1).and.(j.ne.q2).and.(j.ne.q3))then
	        wk1(j,4)=pdh(y,i,wk2(j),type)*bb(i)
	      endif
	    enddo

c perform the limited row reduction.
	    do j=q1+1,n
	      wk1(j,2)=wk1(j,2)-wk1(j,1)*wk1(q1,2)/wk1(q1,1)
	      wk1(j,3)=wk1(j,3)-wk1(j,1)*wk1(q1,3)/wk1(q1,1)
	    enddo
	    do j=q2+1,n
	      wk1(j,3)=wk1(j,3)-wk1(j,2)*wk1(q2,3)/wk1(q2,2)
	    enddo

c place the resulting 3 key diag elements and calculate lnjacob

	    wk1(q1,4)=wk1(q1,1)
	    wk1(q2,4)=wk1(q2,2)
	    wk1(q3,4)=wk1(q3,3)
	    lnjacob(i)=0.d0
	    do j=1,n
	      lnjacob(i)=lnjacob(i)+dlog(abs(wk1(j,4)))
	    enddo
	  enddo

	else
	  i=lam
          lnjacob(i)=0.d0
          medhy=ahch(medy,type,i)
          iqrhy=ahch(y(wk2(q3)),type,i)-ahch(y(wk2(q1)),type,i)
          bb(i)=iqry/iqrhy
          aa(i)=medy-bb(i)*medhy

c fill in non-diagonal elements of the (ordered data) jacobian.
          do j=q1+1,n
            wk1(j,1)=(pdh(y,i,wk2(j),type)*iqry-iqrhy)*
     +      (ahch(y(wk2(j)),type,i)-medhy)/(iqrhy**(2.d0))
            wk1(j,2)=1.d0-pdh(y,i,wk2(j),type)*bb(i)
            wk1(j,3)=(iqrhy-pdh(y,i,wk2(j),type)*iqry)*
     +      (ahch(y(wk2(j)),type,i)-medhy)/(iqrhy**(2.d0))
          enddo

c fill in 3 key diagonal elements
          wk1(q1,1)=(pdh(y,i,wk2(q1),type)*iqry-iqrhy)*
     +    (ahch(y(wk2(q1)),type,i)-medhy)/(iqrhy**(2.d0))+
     +    pdh(y,i,wk2(q1),type)*bb(i)
          wk1(q2,2)=1.d0
          wk1(q3,3)=(iqrhy-pdh(y,i,wk2(q3),type))*
     +    (ahch(y(wk2(q3)),type,i)-medhy)/(iqrhy**(2.d0))+
     +    pdh(y,i,wk2(q3),type)*bb(i)

c fill in remaining diagonal elements
          do j=1,n
            if((j.ne.q1).and.(j.ne.q2).and.(j.ne.q3))then
              wk1(j,4)=pdh(y,i,wk2(j),type)*bb(i)
            endif
          enddo

c perform the limited row reduction.
          do j=q1+1,n
            wk1(j,2)=wk1(j,2)-wk1(j,1)*wk1(q1,2)/wk1(q1,1)
            wk1(j,3)=wk1(j,3)-wk1(j,1)*wk1(q1,3)/wk1(q1,1)
          enddo
          do j=q2+1,n
            wk1(j,3)=wk1(j,3)-wk1(j,2)*wk1(q2,3)/wk1(q2,2)
          enddo

c place the resulting 3 key diag elements and calculate lnjacob

          wk1(q1,4)=wk1(q1,1)
          wk1(q2,4)=wk1(q2,2)
          wk1(q3,4)=wk1(q3,3)
          lnjacob(i)=0.d0
          do j=1,n
            lnjacob(i)=lnjacob(i)+dlog(abs(wk1(j,4)))
          enddo
	
	endif

	return
	end
c_______________________________________________________________
	double precision function pdh(y,i,j,type)

c function returns  dh(y_j)/dy_j with h(.) being of 'type' number i

c routines called: fnorml,mdnris

	integer type,i,j
	double precision y(*)

c local variables

	integer ier
	double precision dummyvar1,dummyvar2,fnorml

	if(type.eq.1)then
	  if(i.eq.1)then
	    pdh=2.d0/(y(j)**3.d0)
	  elseif(i.eq.2)then        
            pdh=1.5d0/(y(j)**2.5d0)
          elseif(i.eq.3)then            
            pdh=1.d0/(y(j)**2.d0)
	  elseif(i.eq.4)then
	    pdh=0.5d0/(y(j)**1.5d0)
          elseif(i.eq.5)then            
            pdh=1.d0/(y(j))
          elseif(i.eq.6)then            
            pdh=-0.5d0/(y(j)**0.5d0)
          elseif(i.eq.7)then            
            pdh=1.d0
          elseif(i.eq.8)then            
            pdh=1.5d0*dsqrt(y(j))
          else            
            pdh=2.d0*y(j)
	  endif
	else
c note here that d(psi^-1(y^lam))/dy = [lam.y^(lam-1)]/[f(psi^-1(u))]
c where u=y^lam and f(x) is the normal density at x.
	  if(i.eq.1)then
	    dummyvar1=y(j)**0.1d0
	    call mdnris(dummyvar1,dummyvar2,ier)
            pdh=0.1d0*(y(j)**(-0.9d0))/fnorml(dummyvar2,0.d0,1.d0)
          elseif(i.eq.2)then 
	    dummyvar1=y(j)**0.25d0
	    call mdnris(dummyvar1,dummyvar2,ier)
            pdh=0.25d0*(y(j)**(-0.75d0))/fnorml(dummyvar2,0.d0,1.d0)
          elseif(i.eq.3)then
	    dummyvar1=y(j)**0.5d0
            call mdnris(dummyvar1,dummyvar2,ier)
            pdh=0.5d0*(y(j)**(-0.5d0))/fnorml(dummyvar2,0.d0,1.d0)
          elseif(i.eq.4)then
	    dummyvar1=y(j)**0.75d0
            call mdnris(dummyvar1,dummyvar2,ier)
            pdh=0.75d0*(y(j)**(-0.25d0))/fnorml(dummyvar2,0.d0,1.d0)
          elseif(i.eq.5)then
	    dummyvar1=y(j)
            call mdnris(dummyvar1,dummyvar2,ier)
            pdh=1.d0/fnorml(dummyvar2,0.d0,1.d0)
          elseif(i.eq.6)then
            dummyvar1=y(j)**1.5d0
            call mdnris(dummyvar1,dummyvar2,ier)
            pdh=1.5d0*(y(j)**0.5d0)/fnorml(dummyvar2,0.d0,1.d0)
          elseif(i.eq.7)then
	    dummyvar1=y(j)**2
	    call mdnris(dummyvar1,dummyvar2,ier)
	    pdh=2.d0*y(j)/fnorml(dummyvar2,0.d0,1.d0)
          elseif(i.eq.8)then
	    pdh=1.d0/(y(j)*(1.d0-y(j)))
          else
            pdh=-1.d0/( dlog(1-y(j))*(1.d0-y(j)) )
          endif
	endif

	return
	end
c________________________________________________________
	double precision function ahch(x,type,i)

c calculates h(x) where h(.) is of type, number i

c routines called: mdnris

	integer type,i
	double precision x

c local variable
        integer ier
	
        if(type.eq.1)then
          if(i.eq.1)then
            ahch=-(x**(-2.d0))
          elseif(i.eq.2)then
            ahch=-(x**(-1.5d0))
          elseif(i.eq.3)then
            ahch=-(x**(-1.d0))
          elseif(i.eq.4)then
            ahch=-(x**(-0.5d0))
          elseif(i.eq.5)then
            ahch=dlog(x)
          elseif(i.eq.6)then
            ahch=x**(0.5d0)
          elseif(i.eq.7)then
            ahch=x
          elseif(i.eq.8)then
            ahch=x**(1.5d0)
          else
            ahch=x**(2.d0)
          endif
        else
          if(i.eq.1)then
            call mdnris(x**0.1d0,ahch,ier)
          elseif(i.eq.2)then
            call mdnris(x**0.25d0,ahch,ier)
          elseif(i.eq.3)then
            call mdnris(x**0.5d0,ahch,ier)
          elseif(i.eq.4)then
            call mdnris(x**0.75d0,ahch,ier)
          elseif(i.eq.5)then
            call mdnris(x,ahch,ier)
          elseif(i.eq.6)then
	    call mdnris(x**1.5d0,ahch,ier)
          elseif(i.eq.7)then
            call mdnris(x**2.d0,ahch,ier)
          elseif(i.eq.8)then
	    ahch=dlog(x/(1.d0-x))
          else
            ahch=dlog(-dlog(1-x))
          endif
	endif
	
	return
	end
c________________________________________________________

	subroutine getpos(a,n,pos)
c   a is a n times 1 vector, pos is an n times 1 vector of 
c integers so that pos(i) is the position of the i-th smallest 
c of the a(i)
c subroutine arguments
         integer n,pos(n)
         double precision a(n)
c routines called: none.
c local variables
         integer i,indxt,ir,j,l
 	double precision q
 
         do 11 j=1,n
           pos(j) = j 
 11	continue
         if( n .eq. 1 ) return 
         l = n/2+1
         ir = n 
 10      continue
           if (l .gt. 1) then 
             l=l-1
             indxt = pos(l)
             q = a(indxt)
           else
             indxt = pos(ir)
             q = a(indxt)
             pos(ir) = pos(1) 
             ir = ir-1
             if (ir .eq. 1) then 
               pos(1) = indxt
               return
             endif
           endif
           i=l 
           j=l+l 
 20        if (j .le. ir) then 
             if (j .lt. ir) then 
               if (a(pos(j)) .lt. a(pos(j+1))) j=j+1
             endif
             if (q .lt. a(pos(j))) then 
               pos(i) = pos(j) 
               i=j 
               j=j+j
             else
               j =ir+1
             endif
             goto 20 
           endif
           pos(i) = indxt
           goto 10 
	   end

           subroutine sort1(a,b,n,pos)
c rearranges elements of the vector a so that
c b(i) = a(pos(i))
c if pos(i) is the position of the i th element of a(i)
c then b is vector a sorted in ascending order .
c
c a and b cannot be the same.
c
c
           integer i,n,pos(*)
           double precision a(*),b(*)
           do 10 i=1,n
              b(i) = a(pos(i))
 10        continue
           return
           end

 

      subroutine newdaxpy(n,da,dx,incx,dy,incy)
c
c     constant times a vector plus a vector.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision dx(1),dy(1),da
      integer i,incx,incy,ix,iy,m,mp1,n
c
      if(n.le.0)return
      if (da .eq. 0.0d0) return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dy(iy) = dy(iy) + da*dx(ix)
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
c
c        code for both increments equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,4)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dy(i) = dy(i) + da*dx(i)
   30 continue
      if( n .lt. 4 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,4
        dy(i) = dy(i) + da*dx(i)
        dy(i + 1) = dy(i + 1) + da*dx(i + 1)
        dy(i + 2) = dy(i + 2) + da*dx(i + 2)
        dy(i + 3) = dy(i + 3) + da*dx(i + 3)
   50 continue
      return
      end

      subroutine  newdswap (n,dx,incx,dy,incy)
c
c     interchanges two vectors.
c     uses unrolled loops for increments equal one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision dx(1),dy(1),dtemp
      integer i,incx,incy,ix,iy,m,mp1,n
c
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c       code for unequal increments or equal increments not equal
c         to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = dx(ix)
        dx(ix) = dy(iy)
        dy(iy) = dtemp
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
c
c       code for both increments equal to 1
c
c
c       clean-up loop
c
   20 m = mod(n,3)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dx(i)
        dx(i) = dy(i)
        dy(i) = dtemp
   30 continue
      if( n .lt. 3 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,3
        dtemp = dx(i)
        dx(i) = dy(i)
        dy(i) = dtemp
        dtemp = dx(i + 1)
        dx(i + 1) = dy(i + 1)
        dy(i + 1) = dtemp
        dtemp = dx(i + 2)
        dx(i + 2) = dy(i + 2)
        dy(i + 2) = dtemp
   50 continue
      return
      end


      subroutine newdchdc(a,lda,p,work,jpvt,job,info)
      integer lda,p,jpvt(1),job,info
      double precision a(lda,1),work(1)
c
c     newdchdc computes the cholesky decomposition of a positive definite
c     matrix.  a pivoting option allows the user to estimate the
c     condition of a positive definite matrix or determine the rank
c     of a positive semidefinite matrix.
c
c     on entry
c
c         a      double precision(lda,p).
c                a contains the matrix whose decomposition is to
c                be computed.  onlt the upper half of a need be stored.
c                the lower part of the array a is not referenced.
c
c         lda    integer.
c                lda is the leading dimension of the array a.
c
c         p      integer.
c                p is the order of the matrix.
c
c         work   double precision.
c                work is a work array.
c
c         jpvt   integer(p).
c                jpvt contains integers that control the selection
c                of the pivot elements, if pivoting has been requested.
c                each diagonal element a(k,k)
c                is placed in one of three classes according to the
c                value of jpvt(k).
c
c                   if jpvt(k) .gt. 0, then x(k) is an initial
c                                      element.
c
c                   if jpvt(k) .eq. 0, then x(k) is a free element.
c
c                   if jpvt(k) .lt. 0, then x(k) is a final element.
c
c                before the decomposition is computed, initial elements
c                are moved by symmetric row and column interchanges to
c                the beginning of the array a and final
c                elements to the end.  both initial and final elements
c                are frozen in place during the computation and only
c                free elements are moved.  at the k-th stage of the
c                reduction, if a(k,k) is occupied by a free element
c                it is interchanged with the largest free element
c                a(l,l) with l .ge. k.  jpvt is not referenced if
c                job .eq. 0.
c
c        job     integer.
c                job is an integer that initiates column pivoting.
c                if job .eq. 0, no pivoting is done.
c                if job .ne. 0, pivoting is done.
c
c     on return
c
c         a      a contains in its upper half the cholesky factor
c                of the matrix a as it has been permuted by pivoting.
c
c         jpvt   jpvt(j) contains the index of the diagonal element
c                of a that was moved into the j-th position,
c                provided pivoting was requested.
c
c         info   contains the index of the last positive diagonal
c                element of the cholesky factor.
c
c     for positive definite matrices info = p is the normal return.
c     for pivoting with positive semidefinite matrices info will
c     in general be less than p.  however, info may be greater than
c     the rank of a, since rounding error can cause an otherwise zero
c     element to be positive. indefinite systems will always cause
c     info to be less than p.
c
c     linpack. this version dated 08/14/78 .
c     j.j. dongarra and g.w. stewart, argonne national laboratory and
c     university of maryland.
c
c
c     blas newdaxpy,newdswap
c     fortran dsqrt
c
c     internal variables
c
      integer pu,pl,plp1,i,j,jp,jt,k,kb,km1,kp1,l,maxl
      double precision temp
      double precision maxdia
      logical swapk,negk
c
      pl = 1
      pu = 0
      info = p
      if (job .eq. 0) go to 160
c
c        pivoting has been requested. rearrange the
c        the elements according to jpvt.
c
         do 70 k = 1, p
            swapk = jpvt(k) .gt. 0
            negk = jpvt(k) .lt. 0
            jpvt(k) = k
            if (negk) jpvt(k) = -jpvt(k)
            if (.not.swapk) go to 60
               if (k .eq. pl) go to 50
                  call newdswap(pl-1,a(1,k),1,a(1,pl),1)
                  temp = a(k,k)
                  a(k,k) = a(pl,pl)
                  a(pl,pl) = temp
                  plp1 = pl + 1
                  if (p .lt. plp1) go to 40
                  do 30 j = plp1, p
                     if (j .ge. k) go to 10
                        temp = a(pl,j)
                        a(pl,j) = a(j,k)
                        a(j,k) = temp
                     go to 20
   10                continue
                     if (j .eq. k) go to 20
                        temp = a(k,j)
                        a(k,j) = a(pl,j)
                        a(pl,j) = temp
   20                continue
   30             continue
   40             continue
                  jpvt(k) = jpvt(pl)
                  jpvt(pl) = k
   50          continue
               pl = pl + 1
   60       continue
   70    continue
         pu = p
         if (p .lt. pl) go to 150
         do 140 kb = pl, p
            k = p - kb + pl
            if (jpvt(k) .ge. 0) go to 130
               jpvt(k) = -jpvt(k)
               if (pu .eq. k) go to 120
                  call newdswap(k-1,a(1,k),1,a(1,pu),1)
                  temp = a(k,k)
                  a(k,k) = a(pu,pu)
                  a(pu,pu) = temp
                  kp1 = k + 1
                  if (p .lt. kp1) go to 110
                  do 100 j = kp1, p
                     if (j .ge. pu) go to 80
                        temp = a(k,j)
                        a(k,j) = a(j,pu)
                        a(j,pu) = temp
                     go to 90
   80                continue
                     if (j .eq. pu) go to 90
                        temp = a(k,j)
                        a(k,j) = a(pu,j)
                        a(pu,j) = temp
   90                continue
  100             continue
  110             continue
                  jt = jpvt(k)
                  jpvt(k) = jpvt(pu)
                  jpvt(pu) = jt
  120          continue
               pu = pu - 1
  130       continue
  140    continue
  150    continue
  160 continue
      do 270 k = 1, p
c
c        reduction loop.
c
         maxdia = a(k,k)
         kp1 = k + 1
         maxl = k
c
c        determine the pivot element.
c
         if (k .lt. pl .or. k .ge. pu) go to 190
            do 180 l = kp1, pu
               if (a(l,l) .gt. maxdia) go to 170
                  maxdia = a(l,l)
                  maxl = l
  170          continue
  180       continue
  190    continue
c
c        quit if the pivot element is not positive.
c
         if (maxdia .gt. 0.0d0) go to 200
            info = k - 1
c     ......exit
            go to 280
  200    continue
         if (k .eq. maxl) go to 210
c
c           start the pivoting and update jpvt.
c
            km1 = k - 1
            call newdswap(km1,a(1,k),1,a(1,maxl),1)
            a(maxl,maxl) = a(k,k)
            a(k,k) = maxdia
            jp = jpvt(maxl)
            jpvt(maxl) = jpvt(k)
            jpvt(k) = jp
  210    continue
c
c        reduction step. pivoting is contained across the rows.
c
         work(k) = dsqrt(a(k,k))
         a(k,k) = work(k)
         if (p .lt. kp1) go to 260
         do 250 j = kp1, p
            if (k .eq. maxl) go to 240
               if (j .ge. maxl) go to 220
                  temp = a(k,j)
                  a(k,j) = a(j,maxl)
                  a(j,maxl) = temp
               go to 230
  220          continue
               if (j .eq. maxl) go to 230
                  temp = a(k,j)
                  a(k,j) = a(maxl,j)
                  a(maxl,j) = temp
  230          continue
  240       continue
            a(k,j) = a(k,j)/work(k)
            work(j) = a(k,j)
            temp = -a(k,j)
            call newdaxpy(j-k,temp,work(kp1),1,a(kp1,j),1)
  250    continue
  260    continue
  270 continue
  280 continue
      return
      end
c----------------------------------------------------------

      subroutine dchex(r,ldr,p,k,l,z,ldz,nz,c,s,job)
      integer ldr,p,k,l,ldz,nz,job
      double precision r(ldr,1),z(ldz,1),s(1)
      double precision c(1)
c
c     dchex updates the cholesky factorization
c
c                   a = trans(r)*r
c
c     of a positive definite matrix a of order p under diagonal
c     permutations of the form
c
c                   trans(e)*a*e
c
c     where e is a permutation matrix.  specifically, given
c     an upper triangular matrix r and a permutation matrix
c     e (which is specified by k, l, and job), dchex determines
c     a orthogonal matrix u such that
c
c                           u*r*e = rr,
c
c     where rr is upper triangular.  at the users option, the
c     transformation u will be multiplied into the array z.
c     if a = trans(x)*x, so that r is the triangular part of the
c     qr factorization of x, then rr is the triangular part of the
c     qr factorization of x*e, i.e. x with its columns permuted.
c     for a less terse description of what dchex does and how
c     it may be applied, see the linpack guide.
c
c     the matrix q is determined as the product u(l-k)*...*u(1)
c     of plane rotations of the form
c
c                           (    c(i)       s(i) )
c                           (                    ) ,
c                           (    -s(i)      c(i) )
c
c     where c(i) is double precision, the rows these rotations operate
c     on are described below.
c
c     there are two types of permutations, which are determined
c     by the value of job.
c
c     1. right circular shift (job = 1).
c
c         the columns are rearranged in the following order.
c
c                1,...,k-1,l,k,k+1,...,l-1,l+1,...,p.
c
c         u is the product of l-k rotations u(i), where u(i)
c         acts in the (l-i,l-i+1)-plane.
c
c     2. left circular shift (job = 2).
c         the columns are rearranged in the following order
c
c                1,...,k-1,k+1,k+2,...,l,k,l+1,...,p.
c
c         u is the product of l-k rotations u(i), where u(i)
c         acts in the (k+i-1,k+i)-plane.
c
c     on entry
c
c         r      double precision(ldr,p), where ldr.ge.p.
c                r contains the upper triangular factor
c                that is to be updated.  elements of r
c                below the diagonal are not referenced.
c
c         ldr    integer.
c                ldr is the leading dimension of the array r.
c
c         p      integer.
c                p is the order of the matrix r.
c
c         k      integer.
c                k is the first column to be permuted.
c
c         l      integer.
c                l is the last column to be permuted.
c                l must be strictly greater than k.
c
c         z      double precision(ldz,nz), where ldz.ge.p.
c                z is an array of nz p-vectors into which the
c                transformation u is multiplied.  z is
c                not referenced if nz = 0.
c
c         ldz    integer.
c                ldz is the leading dimension of the array z.
c
c         nz     integer.
c                nz is the number of columns of the matrix z.
c
c         job    integer.
c                job determines the type of permutation.
c                       job = 1  right circular shift.
c                       job = 2  left circular shift.
c
c     on return
c
c         r      contains the updated factor.
c
c         z      contains the updated matrix z.
c
c         c      double precision(p).
c                c contains the cosines of the transforming rotations.
c
c         s      double precision(p).
c                s contains the sines of the transforming rotations.
c
c     linpack. this version dated 08/14/78 .
c     g.w. stewart, university of maryland, argonne national lab.
c
c     dchex uses the following functions and subroutines.
c
c     blas newdrotg
c     fortran min0
c
      integer i,ii,il,iu,j,jj,km1,kp1,lmk,lm1
      double precision rjp1j,t
c
c     initialize
c
      km1 = k - 1
      kp1 = k + 1
      lmk = l - k
      lm1 = l - 1
c
c     perform the appropriate task.
c
      go to (10,130), job
c
c     right circular shift.
c
   10 continue
c
c        reorder the columns.
c
         do 20 i = 1, l
            ii = l - i + 1
            s(i) = r(ii,l)
   20    continue
         do 40 jj = k, lm1
            j = lm1 - jj + k
            do 30 i = 1, j
               r(i,j+1) = r(i,j)
   30       continue
            r(j+1,j+1) = 0.0d0
   40    continue
         if (k .eq. 1) go to 60
            do 50 i = 1, km1
               ii = l - i + 1
               r(i,k) = s(ii)
   50       continue
   60    continue
c
c        calculate the rotations.
c
         t = s(1)
         do 70 i = 1, lmk
            call newdrotg(s(i+1),t,c(i),s(i))
            t = s(i+1)
   70    continue
         r(k,k) = t
         do 90 j = kp1, p
            il = max0(1,l-j+1)
            do 80 ii = il, lmk
               i = l - ii
               t = c(ii)*r(i,j) + s(ii)*r(i+1,j)
               r(i+1,j) = c(ii)*r(i+1,j) - s(ii)*r(i,j)
               r(i,j) = t
   80       continue
   90    continue
c
c        if required, apply the transformations to z.
c
         if (nz .lt. 1) go to 120
         do 110 j = 1, nz
            do 100 ii = 1, lmk
               i = l - ii
               t = c(ii)*z(i,j) + s(ii)*z(i+1,j)
               z(i+1,j) = c(ii)*z(i+1,j) - s(ii)*z(i,j)
               z(i,j) = t
  100       continue
  110    continue
  120    continue
      go to 260
c
c     left circular shift
c
  130 continue
c
c        reorder the columns
c
         do 140 i = 1, k
            ii = lmk + i
            s(ii) = r(i,k)
  140    continue
         do 160 j = k, lm1
            do 150 i = 1, j
               r(i,j) = r(i,j+1)
  150       continue
            jj = j - km1
            s(jj) = r(j+1,j+1)
  160    continue
         do 170 i = 1, k
            ii = lmk + i
            r(i,l) = s(ii)
  170    continue
         do 180 i = kp1, l
            r(i,l) = 0.0d0
  180    continue
c
c        reduction loop.
c
         do 220 j = k, p
            if (j .eq. k) go to 200
c
c              apply the rotations.
c
               iu = min0(j-1,l-1)
               do 190 i = k, iu
                  ii = i - k + 1
                  t = c(ii)*r(i,j) + s(ii)*r(i+1,j)
                  r(i+1,j) = c(ii)*r(i+1,j) - s(ii)*r(i,j)
                  r(i,j) = t
  190          continue
  200       continue
            if (j .ge. l) go to 210
               jj = j - k + 1
               t = s(jj)
               call newdrotg(r(j,j),t,c(jj),s(jj))
  210       continue
  220    continue
c
c        apply the rotations to z.
c
         if (nz .lt. 1) go to 250
         do 240 j = 1, nz
            do 230 i = k, lm1
               ii = i - km1
               t = c(ii)*z(i,j) + s(ii)*z(i+1,j)
               z(i+1,j) = c(ii)*z(i+1,j) - s(ii)*z(i,j)
               z(i,j) = t
  230       continue
  240    continue
  250    continue
  260 continue
      return
      end

c
c
c
      subroutine newdrotg(da,db,c,s)
c
c     construct givens plane rotation.
c     jack dongarra, linpack, 3/11/78.
c                    modified 9/27/86.
c
      double precision da,db,c,s,roe,scale,r,z
c
      roe = db
      if( dabs(da) .gt. dabs(db) ) roe = da
      scale = dabs(da) + dabs(db)
      if( scale .ne. 0.0d0 ) go to 10
         c = 1.0d0
         s = 0.0d0
         r = 0.0d0
         go to 20
   10 r = scale*dsqrt((da/scale)**2 + (db/scale)**2)
      r = dsign(1.0d0,roe)*r
      c = da/r
      s = db/r
   20 z = s
      if( dabs(c) .gt. 0.0d0 .and. dabs(c) .le. s ) z = 1.0d0/c
      da = r
      db = z
      return
      end

c____________________________________________________

      subroutine newdchdd(r,ldr,p,x,z,ldz,nz,y,rho,c,s,info)
      integer ldr,p,ldz,nz,info
      double precision r(ldr,1),x(1),z(ldz,1),y(1),s(1)
      double precision rho(1),c(1)
c
c     dchdd downdates an augmented cholesky decomposition or the
c     triangular factor of an augmented qr decomposition.
c     specifically, given an upper triangular matrix r of order p,  a
c     row vector x, a column vector z, and a scalar y, dchdd
c     determineds a orthogonal matrix u and a scalar zeta such that
c
c                        (r   z )     (rr  zz)
c                    u * (      )  =  (      ) ,
c                        (0 zeta)     ( x   y)
c
c     where rr is upper triangular.  if r and z have been obtained
c     from the factorization of a least squares problem, then
c     rr and zz are the factors corresponding to the problem
c     with the observation (x,y) removed.  in this case, if rho
c     is the norm of the residual vector, then the norm of
c     the residual vector of the downdated problem is
c     dsqrt(rho**2 - zeta**2). dchdd will simultaneously downdate
c     several triplets (z,y,rho) along with r.
c     for a less terse description of what dchdd does and how
c     it may be applied, see the linpack guide.
c
c     the matrix u is determined as the product u(1)*...*u(p)
c     where u(i) is a rotation in the (p+1,i)-plane of the
c     form
c
c                       ( c(i)     -s(i)     )
c                       (                    ) .
c                       ( s(i)       c(i)    )
c
c     the rotations are chosen so that c(i) is double precision.
c
c     the user is warned that a given downdating problem may
c     be impossible to accomplish or may produce
c     inaccurate results.  for example, this can happen
c     if x is near a vector whose removal will reduce the
c     rank of r.  beware.
c
c     on entry
c
c         r      double precision(ldr,p), where ldr .ge. p.
c                r contains the upper triangular matrix
c                that is to be downdated.  the part of  r
c                below the diagonal is not referenced.
c
c         ldr    integer.
c                ldr is the leading dimension fo the array r.
c
c         p      integer.
c                p is the order of the matrix r.
c
c         x      double precision(p).
c                x contains the row vector that is to
c                be removed from r.  x is not altered by dchdd.
c
c         z      double precision(ldz,nz), where ldz .ge. p.
c                z is an array of nz p-vectors which
c                are to be downdated along with r.
c
c         ldz    integer.
c                ldz is the leading dimension of the array z.
c
c         nz     integer.
c                nz is the number of vectors to be downdated
c                nz may be zero, in which case z, y, and rho
c                are not referenced.
c
c         y      double precision(nz).
c                y contains the scalars for the downdating
c                of the vectors z.  y is not altered by dchdd.
c
c         rho    double precision(nz).
c                rho contains the norms of the residual
c                vectors that are to be downdated.
c
c     on return
c
c         r
c         z      contain the downdated quantities.
c         rho
c
c         c      double precision(p).
c                c contains the cosines of the transforming
c                rotations.
c
c         s      double precision(p).
c                s contains the sines of the transforming
c                rotations.
c
c         info   integer.
c                info is set as follows.
c
c                   info = 0  if the entire downdating
c                             was successful.
c
c                   info =-1  if r could not be downdated.
c                             in this case, all quantities
c                             are left unaltered.
c
c                   info = 1  if some rho could not be
c                             downdated.  the offending rhos are
c                             set to -1.
c
c     linpack. this version dated 08/14/78 .
c     g.w. stewart, university of maryland, argonne national lab.
c
c     dchdd uses the following functions and subprograms.
c
c     fortran dabs
c     blas newddot, newdnrm2
c
      integer i,ii,j
      double precision a,alpha,azeta,norm,newdnrm2
      double precision newddot,t,zeta,b,xx
c
c     solve the system trans(r)*a = x, placing the result
c     in the array s.
c
      info = 0
      s(1) = x(1)/r(1,1)
      if (p .lt. 2) go to 20
      do 10 j = 2, p
         s(j) = x(j) - newddot(j-1,r(1,j),1,s,1)
         s(j) = s(j)/r(j,j)
   10 continue
   20 continue
      norm = newdnrm2(p,s,1)
      if (norm .lt. 1.0d0) go to 30
         info = -1
      go to 120
   30 continue
         alpha = dsqrt(1.0d0-norm**2)
c
c        determine the transformations.
c
         do 40 ii = 1, p
            i = p - ii + 1
            scale = alpha + dabs(s(i))
            a = alpha/scale
            b = s(i)/scale
            norm = dsqrt(a**2+b**2+0.0d0**2)
            c(i) = a/norm
            s(i) = b/norm
            alpha = scale*norm
   40    continue
c
c        apply the transformations to r.
c
         do 60 j = 1, p
            xx = 0.0d0
            do 50 ii = 1, j
               i = j - ii + 1
               t = c(i)*xx + s(i)*r(i,j)
               r(i,j) = c(i)*r(i,j) - s(i)*xx
               xx = t
   50       continue
   60    continue
c
c        if required, downdate z and rho.
c
         if (nz .lt. 1) go to 110
         do 100 j = 1, nz
            zeta = y(j)
            do 70 i = 1, p
               z(i,j) = (z(i,j) - s(i)*zeta)/c(i)
               zeta = c(i)*zeta - s(i)*z(i,j)
   70       continue
            azeta = dabs(zeta)
            if (azeta .le. rho(j)) go to 80
               info = 1
               rho(j) = -1.0d0
            go to 90
   80       continue
               rho(j) = rho(j)*dsqrt(1.0d0-(azeta/rho(j))**2)
   90       continue
  100    continue
  110    continue
  120 continue
      return
      end
c____________________________________________________

      subroutine newdchud(r,ldr,p,x,z,ldz,nz,y,rho,c,s)
      integer ldr,p,ldz,nz
      double precision rho(1),c(1)
      double precision r(ldr,1),x(1),z(ldz,1),y(1),s(1)
c
c     dchud updates an augmented cholesky decomposition of the
c     triangular part of an augmented qr decomposition.  specifically,
c     given an upper triangular matrix r of order p, a row vector
c     x, a column vector z, and a scalar y, dchud determines a
c     untiary matrix u and a scalar zeta such that
c
c
c                              (r  z)     (rr   zz )
c                         u  * (    )  =  (        ) ,
c                              (x  y)     ( 0  zeta)
c
c     where rr is upper triangular.  if r and z have been
c     obtained from the factorization of a least squares
c     problem, then rr and zz are the factors corresponding to
c     the problem with the observation (x,y) appended.  in this
c     case, if rho is the norm of the residual vector, then the
c     norm of the residual vector of the updated problem is
c     dsqrt(rho**2 + zeta**2).  dchud will simultaneously update
c     several triplets (z,y,rho).
c     for a less terse description of what dchud does and how
c     it may be applied, see the linpack guide.
c
c     the matrix u is determined as the product u(p)*...*u(1),
c     where u(i) is a rotation in the (i,p+1) plane of the
c     form
c
c                       (     c(i)      s(i) )
c                       (                    ) .
c                       (    -s(i)      c(i) )
c
c     the rotations are chosen so that c(i) is double precision.
c
c     on entry
c
c         r      double precision(ldr,p), where ldr .ge. p.
c                r contains the upper triangular matrix
c                that is to be updated.  the part of r
c                below the diagonal is not referenced.
c
c         ldr    integer.
c                ldr is the leading dimension of the array r.
c
c         p      integer.
c                p is the order of the matrix r.
c
c         x      double precision(p).
c                x contains the row to be added to r.  x is
c                not altered by dchud.
c
c         z      double precision(ldz,nz), where ldz .ge. p.
c                z is an array containing nz p-vectors to
c                be updated with r.
c
c         ldz    integer.
c                ldz is the leading dimension of the array z.
c
c         nz     integer.
c                nz is the number of vectors to be updated
c                nz may be zero, in which case z, y, and rho
c                are not referenced.
c
c         y      double precision(nz).
c                y contains the scalars for updating the vectors
c                z.  y is not altered by dchud.
c
c         rho    double precision(nz).
c                rho contains the norms of the residual
c                vectors that are to be updated.  if rho(j)
c                is negative, it is left unaltered.
c
c     on return
c
c         rc
c         rho    contain the updated quantities.
c         z
c
c         c      double precision(p).
c                c contains the cosines of the transforming
c                rotations.
c
c         s      double precision(p).
c                s contains the sines of the transforming
c                rotations.
c
c     linpack. this version dated 08/14/78 .
c     g.w. stewart, university of maryland, argonne national lab.
c
c     dchud uses the following functions and subroutines.
c
c     extended blas newdrotg
c     fortran dsqrt
c
      integer i,j,jm1
      double precision azeta,scale
      double precision t,xj,zeta
c
c     update r.
c
      do 30 j = 1, p
         xj = x(j)
c
c        apply the previous rotations.
c
         jm1 = j - 1
         if (jm1 .lt. 1) go to 20
         do 10 i = 1, jm1
            t = c(i)*r(i,j) + s(i)*xj
            xj = c(i)*xj - s(i)*r(i,j)
            r(i,j) = t
   10    continue
   20    continue
c
c        compute the next rotation.
c
         call newdrotg(r(j,j),xj,c(j),s(j))
   30 continue
c
c     if required, update z and rho.
c
      if (nz .lt. 1) go to 70
      do 60 j = 1, nz
         zeta = y(j)
         do 40 i = 1, p
            t = c(i)*z(i,j) + s(i)*zeta
            zeta = c(i)*zeta - s(i)*z(i,j)
            z(i,j) = t
   40    continue
         azeta = dabs(zeta)
         if (azeta .eq. 0.0d0 .or. rho(j) .lt. 0.0d0) go to 50
            scale = azeta + rho(j)
            rho(j) = scale*dsqrt((azeta/scale)**2+(rho(j)/scale)**2)
   50    continue
   60 continue
   70 continue
      return
      end

      double precision function newddot(n,dx,incx,dy,incy)
c
c     forms the dot product of two vectors.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision dx(1),dy(1),dtemp
      integer i,incx,incy,ix,iy,m,mp1,n
c
      newddot = 0.0d0
      dtemp = 0.0d0
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = dtemp + dx(ix)*dy(iy)
        ix = ix + incx
        iy = iy + incy
   10 continue
      newddot = dtemp
      return
c
c        code for both increments equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dtemp + dx(i)*dy(i)
   30 continue
      if( n .lt. 5 ) go to 60
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dtemp = dtemp + dx(i)*dy(i) + dx(i + 1)*dy(i + 1) +
     *   dx(i + 2)*dy(i + 2) + dx(i + 3)*dy(i + 3) + dx(i + 4)*dy(i + 4)
   50 continue
   60 newddot = dtemp
      return
      end


      double precision function newdnrm2 ( n, dx, incx)
      integer i, incx, ix, j, n, next
      double precision   dx(1), cutlo, cuthi, hitest, sum, xmax,zero,one
      data   zero, one /0.0d0, 1.0d0/
c
c     euclidean norm of the n-vector stored in dx() with storage
c     increment incx .
c     if    n .le. 0 return with result = 0.
c     if n .ge. 1 then incx must be .ge. 1
c
c           c.l.lawson, 1978 jan 08
c     modified to correct failure to update ix, 1/25/92.
c     modified 3/93 to return if incx .le. 0.
c
c     four phase method     using two built-in constants that are
c     hopefully applicable to all machines.
c         cutlo = maximum of  dsqrt(u/eps)  over all known machines.
c         cuthi = minimum of  dsqrt(v)      over all known machines.
c     where
c         eps = smallest no. such that eps + 1. .gt. 1.
c         u   = smallest positive no.   (underflow limit)
c         v   = largest  no.            (overflow  limit)
c
c     brief outline of algorithm..
c
c     phase 1    scans zero components.
c     move to phase 2 when a component is nonzero and .le. cutlo
c     move to phase 3 when a component is .gt. cutlo
c     move to phase 4 when a component is .ge. cuthi/m
c     where m = n for x() real and m = 2*n for complex.
c
c     values for cutlo and cuthi..
c     from the environmental parameters listed in the imsl converter
c     document the limiting values are as follows..
c     cutlo, s.p.   u/eps = 2**(-102) for  honeywell.  close seconds are
c                   univac and dec at 2**(-103)
c                   thus cutlo = 2**(-51) = 4.44089e-16
c     cuthi, s.p.   v = 2**127 for univac, honeywell, and dec.
c                   thus cuthi = 2**(63.5) = 1.30438e19
c     cutlo, d.p.   u/eps = 2**(-67) for honeywell and dec.
c                   thus cutlo = 2**(-33.5) = 8.23181d-11
c     cuthi, d.p.   same as s.p.  cuthi = 1.30438d19
c     data cutlo, cuthi / 8.232d-11,  1.304d19 /
c     data cutlo, cuthi / 4.441e-16,  1.304e19 /
      data cutlo, cuthi / 8.232d-11,  1.304d19 /
c
      if(n .gt. 0 .and. incx.gt.0) go to 10
         newdnrm2  = zero
         go to 300
c
   10 assign 30 to next
      sum = zero
      i = 1
      ix = 1
c                                                 begin main loop
   20    go to next,(30, 50, 70, 110)
   30 if( dabs(dx(i)) .gt. cutlo) go to 85
      assign 50 to next
      xmax = zero
c
c                        phase 1.  sum is zero
c
   50 if( dx(i) .eq. zero) go to 200
      if( dabs(dx(i)) .gt. cutlo) go to 85
c
c                                prepare for phase 2.
      assign 70 to next
      go to 105
c
c                                prepare for phase 4.
c
  100 continue
      ix = j
      assign 110 to next
      sum = (sum / dx(i)) / dx(i)
  105 xmax = dabs(dx(i))
      go to 115
c
c                   phase 2.  sum is small.
c                             scale to avoid destructive underflow.
c
   70 if( dabs(dx(i)) .gt. cutlo ) go to 75
c
c                     common code for phases 2 and 4.
c                     in phase 4 sum is large.  scale to avoid overflow.
c
  110 if( dabs(dx(i)) .le. xmax ) go to 115
         sum = one + sum * (xmax / dx(i))**2
         xmax = dabs(dx(i))
         go to 200
c
  115 sum = sum + (dx(i)/xmax)**2
      go to 200
c
c
c                  prepare for phase 3.
c
   75 sum = (sum * xmax) * xmax
c
c
c     for real or d.p. set hitest = cuthi/n
c     for complex      set hitest = cuthi/(2*n)
c
   85 hitest = cuthi/float( n )
c
c                   phase 3.  sum is mid-range.  no scaling.
c
      do 95 j = ix,n
      if(dabs(dx(i)) .ge. hitest) go to 100
         sum = sum + dx(i)**2
         i = i + incx
   95 continue
      newdnrm2 = dsqrt( sum )
      go to 300
c
  200 continue
      ix = ix + 1
      i = i + incx
      if( ix .le. n ) go to 20
c
c              end of main loop.
c
c              compute square root and adjust for scaling.
c
      newdnrm2 = xmax * dsqrt(sum)
  300 continue
      return
      end
	subroutine newranu(dseed,nr,r) 
c generate nr uniform random numbers r() on (0,1]
	integer nr 
	double precision r(nr)
	double precision dseed

c local variables
	integer i                                   
	double precision d2,d3                                  
	data d2/2147483647.d0/
	data d3/2147483711.d0/
      
	do i=1,nr
	  dseed=dmod(16807.d0*dseed,d2)
	  r(i)=dseed/d3
	enddo
	return
	end
c---------------------------------------------------------
	subroutine mdnris(x,y,ier)

c routine front end to calculate y=Psi^-1(x). Here, y,x are dp
c while ier is an error flag =1 if error, =0 if ok.

c routines called: norppf (from the DATAPAC library.)

        integer ier
        double precision x,y

	if((x.le.0.d0).or.(x.ge.1.d0))then
	  ier=1
	endif

        call norppf(x,y,ier)

	return
	end

c-----------------------------------------------------------------
      SUBROUTINE NORPPF(P,PPF)
C
C     PURPOSE--THIS SUBROUTINE COMPUTES THE PERCENT POINT
C              FUNCTION VALUE FOR THE NORMAL (GAUSSIAN)
C              DISTRIBUTION WITH MEAN = 0 AND STANDARD DEVIATION = 1. 
C              THIS DISTRIBUTION IS DEFINED FOR ALL X AND HAS
C              THE PROBABILITY DENSITY FUNCTION
C              F(X) = (1/SQRT(2*PI))*EXP(-X*X/2). 
C              NOTE THAT THE PERCENT POINT FUNCTION OF A DISTRIBUTION 
C              IS IDENTICALLY THE SAME AS THE INVERSE CUMULATIVE
C              DISTRIBUTION FUNCTION OF THE DISTRIBUTION.
C     INPUT  ARGUMENTS--P      = THE SINGLE PRECISION VALUE 
C                                (BETWEEN 0.0 AND 1.0)
C                                AT WHICH THE PERCENT POINT 
C                                FUNCTION IS TO BE EVALUATED.
C     OUTPUT ARGUMENTS--PPF    = THE SINGLE PRECISION PERCENT
C                                POINT FUNCTION VALUE.
C     OUTPUT--THE SINGLE PRECISION PERCENT POINT
C             FUNCTION VALUE PPF.
C     PRINTING--NONE UNLESS AN INPUT ARGUMENT ERROR CONDITION EXISTS. 
C     RESTRICTIONS--P SHOULD BE BETWEEN 0.0 AND 1.0, EXCLUSIVELY.
C     OTHER DATAPAC   SUBROUTINES NEEDED--NONE.
C     FORTRAN LIBRARY SUBROUTINES NEEDED--SQRT, ALOG.
C     MODE OF INTERNAL OPERATIONS--SINGLE PRECISION.
C     LANGUAGE--ANSI FORTRAN. 
C     REFERENCES--ODEH AND EVANS, THE PERCENTAGE POINTS
C                 OF THE NORMAL DISTRIBUTION, ALGORTIHM 70, 
C                 APPLIED STATISTICS, 1974, PAGES 96-97.
C               --EVANS, ALGORITHMS FOR MINIMAL DEGREE
C                 POLYNOMIAL AND RATIONAL APPROXIMATION,
C                 M. SC. THESIS, 1972, UNIVERSITY 
C                 OF VICTORIA, B. C., CANADA.
C               --HASTINGS, APPROXIMATIONS FOR DIGITAL
C                 COMPUTERS, 1955, PAGES 113, 191, 192.
C               --NATIONAL BUREAU OF STANDARDS APPLIED MATHEMATICS
C                 SERIES 55, 1964, PAGE 933, FORMULA 26.2.23.
C               --FILLIBEN, SIMPLE AND ROBUST LINEAR ESTIMATION
C                 OF THE LOCATION PARAMETER OF A SYMMETRIC
C                 DISTRIBUTION (UNPUBLISHED PH.D. DISSERTATION,
C                 PRINCETON UNIVERSITY), 1969, PAGES 21-44, 229-231.
C               --FILLIBEN, 'THE PERCENT POINT FUNCTION',
C                 (UNPUBLISHED MANUSCRIPT), 1970, PAGES 28-31.
C               --JOHNSON AND KOTZ, CONTINUOUS UNIVARIATE
C                 DISTRIBUTIONS--1, 1970, PAGES 40-111.
C               --THE KELLEY STATISTICAL TABLES, 1948.
C               --OWEN, HANDBOOK OF STATISTICAL TABLES,
C                 1962, PAGES 3-16.
C               --PEARSON AND HARTLEY, BIOMETRIKA TABLES
C                 FOR STATISTICIANS, VOLUME 1, 1954,
C                 PAGES 104-113.
C     COMMENTS--THE CODING AS PRESENTED BELOW
C               IS ESSENTIALLY IDENTICAL TO THAT
C               PRESENTED BY ODEH AND EVANS
C               AS ALGORTIHM 70 OF APPLIED STATISTICS.
C               THE PRESENT AUTHOR HAS MODIFIED THE
C               ORIGINAL ODEH AND EVANS CODE WITH ONLY
C               MINOR STYLISTIC CHANGES.
C             --AS POINTED OUT BY ODEH AND EVANS
C               IN APPLIED STATISTICS,
C               THEIR ALGORITHM REPRESENTES A
C               SUBSTANTIAL IMPROVEMENT OVER THE
C               PREVIOUSLY EMPLOYED
C               HASTINGS APPROXIMATION FOR THE
C               NORMAL PERCENT POINT FUNCTION--
C               THE ACCURACY OF APPROXIMATION
C               BEING IMPROVED FROM 4.5*(10**-4)
C               TO 1.5*(10**-8).
C     WRITTEN BY--JAMES J. FILLIBEN
C                 STATISTICAL ENGINEERING LABORATORY (205.03)
C                 NATIONAL BUREAU OF STANDARDS
C                 WASHINGTON, D. C. 20234
C                 PHONE:  301-921-2315
C     ORIGINAL VERSION--JUNE      1972. 
C     UPDATED         --SEPTEMBER 1975. 
C     UPDATED         --NOVEMBER  1975. 
C     UPDATED         --OCTOBER   1976. 
C
C---------------------------------------------------------------------
C
	double precision p,ppf
      DATA P0,P1,P2,P3,P4
     1/-.322232431088,-1.0,
     1 -.342242088547,-.204231210245E-1,
     1 -.453642210148E-4/
      DATA Q0,Q1,Q2,Q3,Q4
     1/.993484626060E-1,.588581570495,
     1 .531103462366,.103537752850,
     1 .38560700634E-2/
C
      IPR=6
C
C     CHECK THE INPUT ARGUMENTS FOR ERRORS
C
c      IF(P.LE.0.0.OR.P.GE.1.0)GOTO50
      GOTO90
c   50 WRITE(IPR,1)
c      WRITE(IPR,46)P
c      RETURN
   90 CONTINUE
c    1 FORMAT(1H ,115H***** FATAL ERROR--THE FIRST  INPUT ARGUMENT TO THE
c     1 NORPPF SUBROUTINE IS OUTSIDE THE ALLOWABLE (0,1) INTERVAL *****)
c   46 FORMAT(1H , 35H***** THE VALUE OF THE ARGUMENT IS ,E15.8,6H *****)
C
C-----START POINT-----------------------------------------------------
C
      IF(P.NE.0.5)GOTO150
      PPF=0.0
      RETURN
C
  150 R=P 
      IF(P.GT.0.5)R=1.0-R
      T=SQRT(-2.0*ALOG(R))
      ANUM=((((T*P4+P3)*T+P2)*T+P1)*T+P0)
      ADEN=((((T*Q4+Q3)*T+Q2)*T+Q1)*T+Q0)
      PPF=T+(ANUM/ADEN)
      IF(P.LT.0.5)PPF=-PPF
      RETURN
C
      END 
